import os
import glob
import math
import random
import argparse
import numpy as np
import seaborn as sns
from Bio import SeqIO
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.patches import Patch


def just_plot_backup(num_list_depth, num_list_gc, ctg_len_list, ctg_color_list, bin_file_list, legend_elements, output_plot):

    num_array_depth = np.array(num_list_depth)
    num_array_gc = np.array(num_list_gc)
    num_array_len = np.array(ctg_len_list)
    fig, ax = plt.subplots()
    scatter = plt.scatter(num_array_depth, num_array_gc, c=ctg_color_list, s=num_array_len, alpha=0.3, linewidths=0)
    plt.title('Contig length (Kbp): %s - %s' % (min(ctg_len_list), max(ctg_len_list)), fontsize=9)
    plt.xlabel("Depth (X)", size=9)
    plt.ylabel("GC (%)", size=9)

    # add size legend
    # kw = dict(prop="sizes", num=6, color='grey', fmt="{x:.2f}", func=lambda s: np.sqrt(s/.3)/3)
    # handles, labels = scatter.legend_elements(**kw, alpha=0.5)
    # legend2 = ax.legend(handles, labels, loc="upper left", title="Size", prop={'size': 5}, bbox_to_anchor=(1, 1))
    # ax.add_artist(legend2)

    # add color legend
    plt.legend(handles=legend_elements, title='MAG', loc='upper left', prop={'size': 6}, bbox_to_anchor=(1, 0.5))

    # save and close
    plt.tight_layout()
    plt.savefig(output_plot)
    plt.close()


num_list_depth  = []
num_list_gc     = []
ctg_len_list    = []
ctg_color_list  = ['green']
bin_file_list   = ['bin1.fa']
output_plot     = '/Users/songweizhi/Desktop/888/plot.pdf'


pwd_each_mag    = '/Users/songweizhi/Desktop/888/4.fa'
depth_file      = '/Users/songweizhi/Desktop/888/filtered_stats_Maxbin_nonEukaryota.txt'


# read in depth info
ctg_depth_dict = dict()
for each_ctg in open(depth_file):
    each_ctg_split = each_ctg.strip().split('\t')
    ctg_depth_dict[each_ctg_split[0]] = float(each_ctg_split[1])

for each_ctg in SeqIO.parse(pwd_each_mag, 'fasta'):
    ctg_depth = ctg_depth_dict[each_ctg.id]
    ctg_seq = str(each_ctg.seq).upper()
    ctg_gc = ((ctg_seq.count('G')) + (ctg_seq.count('C'))) * 100 / len(ctg_seq)
    num_list_depth.append(ctg_depth)
    num_list_gc.append(ctg_gc)
    ctg_len_kbp = len(ctg_seq)/(1024)
    ctg_len_kbp = float("{0:.2f}".format(ctg_len_kbp))
    ctg_len_list.append(ctg_len_kbp)

legend_elements = []
legend_elements.append(Patch(facecolor='green', edgecolor='green', label='bin1.fa'))

just_plot(num_list_depth, num_list_gc, ctg_len_list, ctg_color_list, bin_file_list, legend_elements, output_plot)





import urllib.request
req = urllib.request.Request(url='https://localhost/cgi-bin/test.cgi',
                      data=b'This data is passed to stdin of the CGI')
with urllib.request.urlopen(req) as f:
    print(f.read().decode('utf-8'))
    


import argparse
from ete3 import Tree


rename_leaves_usage = '''
==================== rename_leaves example commands ====================

BioSAK rename_leaves -tree input.tree -txt rename.txt -out output.tree

# format of rename.txt (tab separated)
leaf_1  leaf_1_new_name
leaf_2  leaf_2_new_name

========================================================================
'''


def rename_leaves(args):

    tree_file_in      = args['tree']
    rename_file       = args['txt']
    tree_format       = args['fmt']
    tree_file_out     = args['out']

    mag_rename_dict = {}
    for each_mag in open(rename_file):
        each_mag_split = each_mag.strip().split('\t')
        before_rename = each_mag_split[0]
        after_rename = each_mag_split[1]
        mag_rename_dict[before_rename] = after_rename

    t = Tree(tree_file_in, format=tree_format)

    input_tree_leaf_name_list = []
    for leaf in t:
        input_tree_leaf_name_list.append(leaf.name)

    leaves_with_new_name = 0
    leaves_without_new_name = 0
    for each_raw_name in input_tree_leaf_name_list:
        if each_raw_name in mag_rename_dict:
            leaves_with_new_name += 1
        else:
            leaves_without_new_name += 1

    if leaves_with_new_name == 0:
        print('No leaf on input tree found in rename file, please double check!')
        exit()
    elif leaves_without_new_name > 0:
        print('Found %s leaves in input tree, %s of them were found in the rename file.' % (len(input_tree_leaf_name_list), leaves_with_new_name))

    for leaf in t:
        leaf_name_new = mag_rename_dict.get(leaf.name, leaf.name)
        leaf.name = leaf_name_new
    t.write(format=tree_format, outfile=tree_file_out)

    print('Done!')


if __name__ == '__main__':

    rename_leaves_parser = argparse.ArgumentParser()
    rename_leaves_parser.add_argument('-tree', required=True,             help='input tree')
    rename_leaves_parser.add_argument('-txt',  required=True,             help='rename file')
    rename_leaves_parser.add_argument('-fmt',  required=False, default=1, help='tree format, default: 1')
    rename_leaves_parser.add_argument('-out',  required=True,             help='output tree')
    args = vars(rename_leaves_parser.parse_args())
    rename_leaves(args)



import os
import argparse
from time import sleep
from datetime import datetime


get_Pfam_hmms_usage = '''
===================================== get_Pfam_hmms example commands =====================================

# module needed
module load hmmer/3.2.1

# for completed genome
BioSAK get_Pfam_hmms -pfam Pfam-A.hmm -id needed_ids.txt

==========================================================================================================
'''


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def get_Pfam_hmms(args):

    downloaded_pfam_db = args['pfam']
    hmm_id_file = args['id']

    # define file name
    time_format = '[%Y-%m-%d %H:%M:%S]'
    pfam_db_path, pfam_db_basename, pfam_db_extension =             sep_path_basename_ext(downloaded_pfam_db)
    hmm_id_file_path, hmm_id_file_basename, hmm_id_file_extension = sep_path_basename_ext(hmm_id_file)
    downloaded_pfam_db_stats =  '%s/%s.stat'                        % (pfam_db_path, pfam_db_basename)
    Pfam_41_ids_updated =       '%s/%s_latest_version.txt'          % (hmm_id_file_path, hmm_id_file_basename)
    Pfam_profiles_updated =     '%s/%s.hmm'                         % (hmm_id_file_path, hmm_id_file_basename)

    # read in 41 Pfam ids
    sleep(0.5)
    print('%s %s' % ((datetime.now().strftime(time_format)), 'Read in provided profile ids'))
    Pfam_41_id_list_no_version = []
    for each_hmm in open(hmm_id_file):
        Pfam_41_id_list_no_version.append(each_hmm.strip())

    # get all hmm profile id from downloaded db
    sleep(0.5)
    print('%s %s' % ((datetime.now().strftime(time_format)), 'Get summary statistics for %s' % downloaded_pfam_db))
    hmmstat_cmd_db = 'hmmstat %s > %s' % (downloaded_pfam_db, downloaded_pfam_db_stats)
    os.system(hmmstat_cmd_db)

    # parse Pfam-A.hmm.stat and get updated id file
    print('%s %s' % ((datetime.now().strftime(time_format)), 'Get the latest version of provided profiles'))
    id_to_version_dict_new = {}
    Pfam_41_ids_updated_handle = open(Pfam_41_ids_updated, 'w')
    for each_hmm in open(downloaded_pfam_db_stats):

        if not each_hmm.startswith('#') and (each_hmm.strip() != ''):

            each_hmm_split = each_hmm.strip().split(' ')

            # remove '' from each_hmm_split
            each_hmm_split_no_space = []
            for i in each_hmm_split:
                if i != '':
                    each_hmm_split_no_space.append(i)

            hmm_id_version = each_hmm_split_no_space[2]
            hmm_id_no_version = hmm_id_version.split('.')[0]
            id_to_version_dict_new[hmm_id_no_version] = hmm_id_version

            if hmm_id_no_version in Pfam_41_id_list_no_version:
                Pfam_41_ids_updated_handle.write('%s\n' % hmm_id_version)

    Pfam_41_ids_updated_handle.close()

    # extract updated hmm profile from downloaded db
    sleep(0.5)
    print('%s %s' % ((datetime.now().strftime(time_format)), 'Extract profiles'))
    hmmfetch_cmd = 'hmmfetch -f %s %s > %s' % (downloaded_pfam_db, Pfam_41_ids_updated, Pfam_profiles_updated)
    os.system(hmmfetch_cmd)

    sleep(0.5)
    print('%s %s' % ((datetime.now().strftime(time_format)), 'Delete temporary files'))
    os.system('rm %s' % downloaded_pfam_db_stats)

    sleep(0.5)
    print('%s %s' % ((datetime.now().strftime(time_format)), 'Done!'))


if __name__ == '__main__':

    # initialize the options parser
    parser = argparse.ArgumentParser()

    # arguments for PI
    parser.add_argument('-pfam', required=True, help='Pfam db file, normally with name Pfam-A.hmm')
    parser.add_argument('-id',   required=True, help='ids of profiles need to be extracted, one id per line')

    args = vars(parser.parse_args())

    get_Pfam_hmms(args)



#!/usr/bin/env python
"""
A utility script to split a large FASTA file into smaller ones, with an 
arbitrary number of sequences in each one. 

Takes two arguments:
    1) The large FASTA file that needs to be split
    2) The number of sequences that should be put into each FASTA
"""
#   define a usage message
usage = """Split_FASTA.py - splits a large FASTA file into smaller ones.

Usage:
Split_FASTA.py [FILE] [NUM_SEQS]

Splits [FILE] into a collection of smaller files, with [NUM_SEQS] sequences
in each one."""

#   To take arguments on the command line
import sys
#   To do some math
import math
#   To read/write FASTA. Wrap in a try/except
try:
    from Bio import SeqIO
except ImportError:
    print "This script requires BioPython to be installed!"

#   Simple check to make sure that the correct number of arguments are given
if len(sys.argv) != 3:
    print usage
    exit(1)

#   Set these names so that it is easier to follow them in the code
to_split = sys.argv[1]
num_seqs = sys.argv[2]

#   Make sure that the file provided is readable. 
try:
    open(to_split, 'r')
except IOError:
    print "Error! " + to_split + " does not exist, or is not readable!"
    exit(1)

#   And make sure that the second argument is a positive integer
#   The try-Except block tests for integer, and the if tests for positive
try:
    int(num_seqs)
except ValueError:
    print "Error! Please provide a positive integer!"
    exit(1)
if int(num_seqs) <= 0:
    print "Error! Please provide a positive integer!"
    exit(1)

#   Read in the FASTA file
all_fasta = list(SeqIO.parse(to_split, 'fasta'))
#   How many files to split into?
num_files = int(math.ceil(len(all_fasta)/float(num_seqs)))
#   Print a little message
print "Will split " + to_split + " into " + str(num_files) + " files, with " + str(num_seqs) + " seqs per file."
#   Start splitting!
i = 1
while i <= num_files:
    #   Calculate the start
    start = int(i-1) * int(num_seqs)
    #   and the end
    end = int(i) * int(num_seqs)
    #   Generate the filename
    #   strip off the extension, and get the rest
    filename = to_split.split('.')[:-1][0] + '_' + str(i) + '.fasta'
    #   Write the sequences into the file
    SeqIO.write(all_fasta[start:end], filename, 'fasta')
    #   increment the counter
    i += 1




#!/usr/bin/env python
import os
import argparse
from Bio import SeqIO
from datetime import datetime
from BioSAK.global_functions import time_format


select_seq_usage = '''
======================= select_seq example commands =======================

# Extract sequences with provided id
BioSAK select_seq -seq ctg.fasta -id seq_id.txt -out output.1.fa -option 1 
BioSAK select_seq -seq ctg.fastq -id seq_id.txt -out output.1.fq -option 1 -fq

# Extract sequences except those in seq_id.txt
BioSAK select_seq -seq ctg.fasta -id seq_id.txt -out output.0.fa -option 0
BioSAK select_seq -seq ctg.fasta -id seq_id.txt -out output.0.fa -option 0 -oneline

# seq_id.txt file format: one id per line, great than symbol exculded.

===========================================================================
'''

def select_seq(args):

    # read in argument
    seq_file        = args['seq']
    id_file         = args['id']
    select_option   = args['option']
    output_file     = args['out']
    one_line        = args['oneline']
    in_fastq        = args['fq']

    # get provided id list
    print(datetime.now().strftime(time_format) + 'Read in sequence id in %s' % id_file)
    seq_id_list = set()
    for seq_id in open(id_file):
        seq_id_list.add(seq_id.strip())

    seq_in_format = 'fasta'
    if in_fastq is True:
        seq_in_format = 'fastq'

    # report
    if select_option == 1:
        print(datetime.now().strftime(time_format) + 'Extracting sequences in %s' % id_file)
    if select_option == 0:
        print(datetime.now().strftime(time_format) + 'Extracting sequences except those in %s' % id_file)

    # extract sequences
    output_file_handle = open(output_file, 'w')
    for seq_record in SeqIO.parse(seq_file, seq_in_format):
        seq_id = seq_record.id
        if select_option == 1:
            if seq_id in seq_id_list:

                if in_fastq is False:
                    if one_line is False:
                        SeqIO.write(seq_record, output_file_handle, 'fasta')
                    else:
                        SeqIO.write(seq_record, output_file_handle, 'fasta-2line')
                else:
                    SeqIO.write(seq_record, output_file_handle, 'fastq')

        if select_option == 0:
            if seq_id not in seq_id_list:

                if in_fastq is False:
                    if one_line is False:
                        SeqIO.write(seq_record, output_file_handle, 'fasta')
                    else:
                        SeqIO.write(seq_record, output_file_handle, 'fasta-2line')
                else:
                    SeqIO.write(seq_record, output_file_handle, 'fastq')

    output_file_handle.close()

    print(datetime.now().strftime(time_format) + 'Sequences exported to %s' % output_file)
    print(datetime.now().strftime(time_format) + 'Done!')


if __name__ == '__main__':

    # initialize the options parser
    parser = argparse.ArgumentParser()

    # arguments for select_seq
    parser.add_argument('-seq',       required=True,                        help='sequence file')
    parser.add_argument('-id',        required=True,                        help='sequence ids,one id per line')
    parser.add_argument('-option',    required=True, type=int,              help='choose from 0 and 1')
    parser.add_argument('-out',       required=True,                        help='output file')
    parser.add_argument('-fq',        required=False, action="store_true",  help='in fastq format, default: fa')
    parser.add_argument('-oneline',   required=False, action="store_true",  help='put sequence in single line')
    args = vars(parser.parse_args())
    select_seq(args)



import os
import glob
import shutil
import argparse
import multiprocessing as mp
from datetime import datetime


Prokka_parser_usage = '''
==================== Prokka example commands ====================

# for completed genome
BioSAK Prokka -i gnm_folder -x fa -t 6
BioSAK Prokka -i MAG_folder -x fa -t 6 -meta

==========================================================================
'''


def Prokka(argument_list):

    input_genome = argument_list[0]
    input_genome_folder = argument_list[1]
    meta_mode = argument_list[2]

    # # prepare command (according to Prokka)
    # input_genome_basename, input_genome_ext = os.path.splitext(input_genome)
    # pwd_input_genome = '%s/%s' % (input_genome_folder, input_genome)
    #
    # prodigal_cmd_meta = 'prodigal -f sco -q -c -m -g 11 -p meta -i %s -o %s' % (pwd_input_genome, pwd_output_sco)
    # prodigal_cmd_nonmeta = 'prodigal -f sco -q -c -m -g 11 -i %s -o %s' % (pwd_input_genome, pwd_output_sco)
    #
    # if meta_mode is True:
    #     prodigal_cmd = prodigal_cmd_meta
    # else:
    #     prodigal_cmd = prodigal_cmd_nonmeta
    #
    # os.system(prodigal_cmd)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-i',          required=True,                          help='input genome folder')
    parser.add_argument('-x',          required=True, default='fasta',         help='file extension')
    parser.add_argument('-d',          required=True,                          help='genome domain, Bacteria or Archaea')
    parser.add_argument('-meta',       required=False, action="store_true",    help='annotation mode for metagenome assembled genomes (MAGs)')
    parser.add_argument('-t',          required=False, type=int, default=1,    help='number of threads')
    args = vars(parser.parse_args())
    Prokka(args)


'''
prokka --compliant --cpus 12 --kingdom Bacteria --prefix contig_2560 --locustag contig_2560 --outdir contig_2560_prokka_wd contig_2560.fa
prokka --compliant --cpus 12 --kingdom Archaea --prefix NorthSea_bin022 --locustag NorthSea_bin022 --outdir NorthSea_bin022_prokka_wd NorthSea_bin022.fasta
--metagenome
'''


#!/usr/bin/env python3

import os
import argparse
from BioSAK.global_functions import sep_path_basename_ext


reads2bam_usage = '''
=================================== reads2bam example commands ===================================

module load bowtie/2.3.5.1
module load samtools/1.10
BioSAK reads2bam -p Demo -ref ref.fa -r1 R1.fa -r2 R1.fa -u unpaired.fa -index_ref -t 12
BioSAK reads2bam -p Demo -ref ref.fa -r1 R1.fq -r2 R1.fq -fastq -index_ref -t 12
BioSAK reads2bam -p Demo -ref ref.fa -u unpaired.fa -index_ref -t 12 -tmp -mismatch 2
BioSAK reads2bam -p Demo -ref ref.fa -u unpaired_R1.fa,unpaired_R2.fa -index_ref -t 12 -tmp
BioSAK reads2bam -p Demo -ref ref.fa -u unpaired_R1.fa,unpaired_R2.fa -index_ref -t 12 -tmp -local

==================================================================================================
'''

def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


#!/usr/bin/env python3

# Copyright (C) 2020, Weizhi Song, Torsten Thomas.
# songwz03@gmail.com or t.thomas@unsw.edu.au

# MarkerMAG is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# MarkerMAG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import glob
import shutil
import argparse
import numpy as np
import pandas as pd
import seaborn as sns
from Bio import SeqIO
from Bio.Seq import Seq
import multiprocessing as mp
from datetime import datetime
import plotly.graph_objects as go
from Bio.SeqRecord import SeqRecord
from distutils.spawn import find_executable

link_Marker_MAG_usage = '''
=================================== MarkerMAG example commands ===================================
# example commands
MarkerMAG link -p Test -r1 R1.fasta -r2 R2.fasta -marker 16S_seqs.fa -mag MAG_folder -x fa -t 6
# For more details: https://github.com/songweizhi/MarkerMAG
==================================================================================================
'''


def sam_flag_to_rc(flag_value):

    read_rced = 'na'
    if flag_value != '':
        binary_flag = "{0:b}".format(int(flag_value))
        binary_flag_len = len(str(binary_flag))
        binary_flag_polished = '0' * (12 - binary_flag_len) + str(binary_flag)

        if binary_flag_polished[7] == '0':
            read_rced = False
        if binary_flag_polished[7] == '1':
            read_rced = True

    return read_rced


def get_rc(seq_in):
    seq_in_rc = str(SeqRecord(Seq(seq_in)).reverse_complement().seq)
    return seq_in_rc


def force_create_folder(folder_to_create):
    if os.path.isdir(folder_to_create):
        shutil.rmtree(folder_to_create, ignore_errors=True)
        if os.path.isdir(folder_to_create):
            shutil.rmtree(folder_to_create, ignore_errors=True)
            if os.path.isdir(folder_to_create):
                shutil.rmtree(folder_to_create, ignore_errors=True)
                if os.path.isdir(folder_to_create):
                    shutil.rmtree(folder_to_create, ignore_errors=True)
    os.mkdir(folder_to_create)


def report_and_log(message_for_report, log_file, keep_quiet):

    time_format = '[%Y-%m-%d %H:%M:%S]'
    with open(log_file, 'a') as log_handle:
        log_handle.write('%s %s\n' % ((datetime.now().strftime(time_format)), message_for_report))

    if keep_quiet is False:
        print('%s %s' % ((datetime.now().strftime(time_format)), message_for_report))


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def get_read_num_and_length(reads_file, tmp_file_location, seqtk_exe):

    reads_file_line_num = '%s/R1_line_num.txt'  % (tmp_file_location)
    reads_file_sub1000  = '%s/R1_sub1000.fasta' % (tmp_file_location)

    # get the number of paired reads
    os.system('wc -l %s > %s' % (reads_file, reads_file_line_num))
    paired_reads_num = int(int(open(reads_file_line_num).readline().strip().split(' ')[0]) / 2)
    if reads_file[-1] in ['Q', 'q']:
        paired_reads_num = int(int(open(reads_file_line_num).readline().strip().split(' ')[0])/4)

    # subsample 1000 reads
    os.system('%s sample -s100 %s 1000 > %s' % (seqtk_exe, reads_file, reads_file_sub1000))

    read_len_list = []
    for each_seq in open(reads_file_sub1000):
        if each_seq[0] not in ['>', '@', '+']:
            read_len_list.append(len(each_seq.strip()))

    read_len_median = np.median(read_len_list)
    read_len_max    = np.max(read_len_list)

    os.system('rm %s' % reads_file_line_num)
    os.system('rm %s' % reads_file_sub1000)

    return paired_reads_num, read_len_median, read_len_max


def sep_paired_and_singleton_reads(fasta_in, fasta_out_r1, fasta_out_r2, fasta_out_singleton):
    reads_pair_dict = {}
    for read_record in SeqIO.parse(fasta_in, 'fasta'):
        read_id_base = '.'.join(read_record.id.split('.')[:-1])
        read_strand = read_record.id.split('.')[-1]
        if read_id_base not in reads_pair_dict:
            reads_pair_dict[read_id_base] = {read_strand}
        else:
            reads_pair_dict[read_id_base].add(read_strand)

    read_list_paired = set()
    read_list_singleton = set()
    for read_base in reads_pair_dict:
        if len(reads_pair_dict[read_base]) == 1:
            read_list_singleton.add(read_base)
        if len(reads_pair_dict[read_base]) == 2:
            read_list_paired.add(read_base)

    fasta_out_r1_handle = open(fasta_out_r1, 'w')
    fasta_out_r2_handle = open(fasta_out_r2, 'w')
    fasta_out_singleton_handle = open(fasta_out_singleton, 'w')

    for read_record in SeqIO.parse(fasta_in, 'fasta'):

        read_id_base = '.'.join(read_record.id.split('.')[:-1])
        read_strand = read_record.id.split('.')[-1]

        if read_id_base in read_list_singleton:
            fasta_out_singleton_handle.write('>%s\n' % read_record.id)
            fasta_out_singleton_handle.write('%s\n' % str(read_record.seq))

        if read_id_base in read_list_paired:

            if read_strand == '1':
                fasta_out_r1_handle.write('>%s\n' % read_record.id)
                fasta_out_r1_handle.write('%s\n' % str(read_record.seq))

            if read_strand == '2':
                fasta_out_r2_handle.write('>%s\n' % read_record.id)
                fasta_out_r2_handle.write('%s\n' % str(read_record.seq))

    fasta_out_r1_handle.close()
    fasta_out_r2_handle.close()
    fasta_out_singleton_handle.close()


def remove_clp_in_middle(sam_in, sam_out):

    sam_out_handle = open(sam_out, 'w')

    marker_len_dict = {}
    for each_read in open(sam_in):
        each_read_split = each_read.strip().split('\t')
        if each_read.startswith('@'):
            sam_out_handle.write(each_read)

            marker_id = ''
            marker_len = 0
            for each_element in each_read_split:
                if each_element.startswith('SN:'):
                    marker_id = each_element[3:]
                if each_element.startswith('LN:'):
                    marker_len = int(each_element[3:])
            marker_len_dict[marker_id] = marker_len
        else:
            cigar = each_read_split[5]

            # check if clp in the middle
            if ('S' not in cigar) and ('s' not in cigar):
                sam_out_handle.write(each_read)
            else:
                ref_id = each_read_split[2]
                ref_pos = int(each_read_split[3])
                cigar_splitted = cigar_splitter(cigar)
                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(cigar_splitted)

                clip_in_middle = True
                if (cigar_splitted[0][-1] in ['S', 's']) and (ref_pos == 1):
                    clip_in_middle = False
                if (cigar_splitted[-1][-1] in ['S', 's']):
                    if (ref_pos + r1_aligned_len - 1) == marker_len_dict[ref_id]:
                        clip_in_middle = False

                if clip_in_middle is False:
                    sam_out_handle.write(each_read)

    sam_out_handle.close()


def remove_high_mismatch(sam_in, mismatch_cutoff, sam_out):
    sam_out_handle = open(sam_out, 'w')
    for each_read in open(sam_in):
        each_read_split = each_read.strip().split('\t')
        if each_read.startswith('@'):
            sam_out_handle.write(each_read)
        else:
            cigar = each_read_split[5]
            if cigar == '*':
                sam_out_handle.write(each_read)
            else:
                cigar_splitted = cigar_splitter(cigar)
                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(cigar_splitted)
                if r1_mismatch_pct <= mismatch_cutoff:
                    sam_out_handle.write(each_read)
    sam_out_handle.close()


def get_ctg_mean_depth_by_samtools_coverage(index_ref, ref_seq, reads_r1, reads_r2, reads_unpaired, subsample_rate, num_threads):

    ref_seq_file_path, ref_seq_file_basename, ref_seq_file_extension = sep_path_basename_ext(ref_seq)

    sam_file                                                            = '%s/%s.sam'                                                           % (ref_seq_file_path, ref_seq_file_basename)
    sam_file_one_end_clp                                                = '%s/%s_one_end_clp.sam'                                               % (ref_seq_file_path, ref_seq_file_basename)
    sam_file_one_end_clp_no_middle                                      = '%s/%s_one_end_clp_no_middle.sam'                                     % (ref_seq_file_path, ref_seq_file_basename)
    sam_file_one_end_clp_no_middle_reformatted                          = '%s/%s_one_end_clp_no_middle_reformatted.sam'                         % (ref_seq_file_path, ref_seq_file_basename)
    sam_file_one_end_clp_no_middle_reformatted_log                      = '%s/%s_one_end_clp_no_middle_reformatted.log'                         % (ref_seq_file_path, ref_seq_file_basename)
    sam_file_one_end_clp_no_middle_reformatted_best_match               = '%s/%s_one_end_clp_no_middle_reformatted_best_match.sam'              % (ref_seq_file_path, ref_seq_file_basename)
    sam_file_one_end_clp_no_middle_reformatted_best_match_low_mismatch  = '%s/%s_one_end_clp_no_middle_reformatted_best_match_low_mismatch.sam' % (ref_seq_file_path, ref_seq_file_basename)
    sam_file_sorted                                                     = '%s/%s_sorted.sam'                                                    % (ref_seq_file_path, ref_seq_file_basename)
    coverage_file                                                       = '%s/%s_cov.txt'                                                       % (ref_seq_file_path, ref_seq_file_basename)

    # build reference index
    cmd_bowtie2_build   = 'bowtie2-build --quiet --threads %s -f %s %s/%s' % (num_threads, ref_seq, ref_seq_file_path, ref_seq_file_basename)
    if index_ref is True:
        os.system(cmd_bowtie2_build)

    # mapping
    # if reads_unpaired == '':
    #     cmd_bowtie2_mapping = 'bowtie2 -x %s/%s -1 %s -2 %s -S %s -p %s --all -f --quiet' % (ref_seq_file_path, ref_seq_file_basename, reads_r1, reads_r2, sam_file, num_threads)
    # else:
    #     cmd_bowtie2_mapping = 'bowtie2 -x %s/%s -1 %s -2 %s -U %s -S %s -p %s --all -f --quiet' % (ref_seq_file_path, ref_seq_file_basename, reads_r1, reads_r2, reads_unpaired, sam_file, num_threads)

    # mapping
    if reads_unpaired == '':
        cmd_bowtie2_mapping = 'bowtie2 -x %s/%s -U %s,%s -S %s -p %s --local --all --no-unal -N 1 -L 30 -f --quiet' % (ref_seq_file_path, ref_seq_file_basename, reads_r1, reads_r2, sam_file, num_threads)
    else:
        cmd_bowtie2_mapping = 'bowtie2 -x %s/%s -U %s,%s,%s -S %s -p %s --local --all --no-unal -N 1 -L 30 -f --quiet' % (ref_seq_file_path, ref_seq_file_basename, reads_r1, reads_r2, reads_unpaired, sam_file, num_threads)

    os.system(cmd_bowtie2_mapping)

    # filter mapping
    remove_both_ends_clp(sam_file, sam_file_one_end_clp)
    remove_clp_in_middle(sam_file_one_end_clp, sam_file_one_end_clp_no_middle)
    bbmap_reformat_cmd = 'reformat.sh in=%s out=%s sam=1.4 2> %s' % (sam_file_one_end_clp_no_middle, sam_file_one_end_clp_no_middle_reformatted, sam_file_one_end_clp_no_middle_reformatted_log)
    os.system(bbmap_reformat_cmd)
    keep_best_matches_in_sam_keep_short_M(sam_file_one_end_clp_no_middle_reformatted, 35, sam_file_one_end_clp_no_middle_reformatted_best_match)
    remove_high_mismatch(sam_file_one_end_clp_no_middle_reformatted_best_match, 2, sam_file_one_end_clp_no_middle_reformatted_best_match_low_mismatch)

    # sort mapping
    cmd_samtools_sort = 'samtools sort %s -o %s' % (sam_file_one_end_clp_no_middle_reformatted_best_match_low_mismatch, sam_file_sorted)
    os.system(cmd_samtools_sort)

    # get mean depth
    cmd_samtools_coverage = 'samtools coverage --ff 4 %s -o %s' % (sam_file_sorted, coverage_file)
    os.system(cmd_samtools_coverage)

    # remove sam files
    os.system('rm %s' % sam_file)
    # os.system('rm %s' % sam_file_sorted)

    # store mean depth into dict
    mean_depth_dict_ctg = {}
    ctg_len_dict = {}
    for each_ctg_depth in open(coverage_file):
        if not each_ctg_depth.startswith('#'):
            ctg_depth_split = each_ctg_depth.strip().split('\t')
            ctg_id = ctg_depth_split[0]
            ctg_len = int(ctg_depth_split[2])
            ctg_depth = float(ctg_depth_split[6]) * (1 / subsample_rate)
            mean_depth_dict_ctg[ctg_id] = ctg_depth
            ctg_len_dict[ctg_id] = ctg_len

    return mean_depth_dict_ctg, ctg_len_dict


def remove_reads_with_multi_best_aln(sam_in, sam_out):

    sam_out_tmp = '%s.tmp' % sam_out

    multi_aligned_reads = set()
    best_hit_cigar_dict = {}
    sam_out_best_hits_handle = open(sam_out_tmp, 'w')
    for each_line in open(sam_in):
        if each_line.startswith('@'):
            sam_out_best_hits_handle.write(each_line)
        else:
            each_line_split = each_line.strip().split('\t')
            read_id = each_line_split[0]
            cigar   = each_line_split[5]
            if read_id not in best_hit_cigar_dict:
                best_hit_cigar_dict[read_id] = cigar
                sam_out_best_hits_handle.write(each_line)
            else:
                if cigar == best_hit_cigar_dict[read_id]:
                    sam_out_best_hits_handle.write(each_line)
                    multi_aligned_reads.add(read_id)
    sam_out_best_hits_handle.close()

    sam_out_no_ambiguous_handle = open(sam_out, 'w')
    for best_aln in open(sam_out_tmp):
        if best_aln.startswith('@'):
            sam_out_no_ambiguous_handle.write(best_aln)
        else:
            read_id = best_aln.strip().split('\t')[0]
            if read_id not in multi_aligned_reads:
                sam_out_no_ambiguous_handle.write(best_aln)
    sam_out_no_ambiguous_handle.close()


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def split_list(list_in, subset_num):

    list_in_formatted = [i for i in list_in]

    # get the number of element per subset
    file_num_per_folder = round(len(list_in_formatted) / subset_num)

    n = 1
    lol_out = []
    while n <= subset_num:

        if n < subset_num:
            current_subset_elements = {i for i in list_in_formatted[(file_num_per_folder * (n - 1)):(file_num_per_folder * n)]}
            lol_out.append(current_subset_elements)
        else:
            current_subset_elements = {i for i in list_in_formatted[(file_num_per_folder * (n - 1)):]}
            lol_out.append(current_subset_elements)

        n += 1

    return lol_out


def stats_dict_to_sankey_file_in(clipping_stats_dict, paired_stats_dict, sankey_file_in_clipping, sankey_file_in_paired):

    # prepare input file for plot of clipping mapped reads
    sankey_file_in_clipping_handle = open(sankey_file_in_clipping, 'w')
    sankey_file_in_clipping_handle.write('MarkerGene,GenomicSeq,Number\n')
    for each_clipping in clipping_stats_dict:
        sankey_file_in_clipping_handle.write('%s,%s\n' % (','.join(each_clipping.split('_|_')), clipping_stats_dict[each_clipping]))
    sankey_file_in_clipping_handle.close()

    # prepare input file for plot of paired reads
    sankey_file_in_paired_handle = open(sankey_file_in_paired, 'w')
    sankey_file_in_paired_handle.write('MarkerGene,GenomicSeq,Number\n')
    for each_paired in paired_stats_dict:
        sankey_file_in_paired_handle.write('%s,%s\n' % (','.join(each_paired.split('_|_')), paired_stats_dict[each_paired]))
    sankey_file_in_paired_handle.close()


def sort_csv_by_col(file_in, file_out, col_header):
    df_in        = pd.read_csv(file_in)
    df_in_sorted = df_in.sort_values(by=[col_header], ascending=False)
    df_in_sorted.to_csv(file_out, index=False)


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


def filter_linkages_iteratively_backup(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:

                # consider depth
                if min_16s_gnm_multiple > 0:
                    MarkerGene_depth = marker_gene_depth_dict[MarkerGene]
                    GenomicSeq_depth = genomic_seq_depth_dict[GenomicSeq]
                    if (MarkerGene_depth/GenomicSeq_depth) >= min_16s_gnm_multiple:
                        if MarkerGene not in MarkerGene_with_assignment:

                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                                iden_with_best_marker = 0
                                if key_str in pairwise_16s_iden_dict:
                                    iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                # ignore depth
                else:
                    if MarkerGene not in MarkerGene_with_assignment:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                            iden_with_best_marker = 0
                            if key_str in pairwise_16s_iden_dict:
                                iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)

    file_out_handle.close()

    # remove tmp file
    # os.remove(file_in_sorted)


def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, within_gnm_linkage_num_diff, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    gnm_max_link_num_dict = {}
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:
                if MarkerGene not in MarkerGene_with_assignment:

                    # consider depth
                    if min_16s_gnm_multiple > 0:

                        # get marker and genome depth
                        MarkerGene_depth = marker_gene_depth_dict.get(MarkerGene, 'na')
                        GenomicSeq_depth = genomic_seq_depth_dict.get(GenomicSeq, 'na')
                        marker_genome_depth_ratio = 'na'
                        if (MarkerGene_depth != 'na') and (GenomicSeq_depth != 'na'):
                            if GenomicSeq_depth > 0:
                                marker_genome_depth_ratio = MarkerGene_depth / GenomicSeq_depth

                        if marker_genome_depth_ratio >= min_16s_gnm_multiple:
                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                gnm_max_link_num_dict[GenomicSeq] = linkage_num
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                # get identity with best marker
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                                iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                    if (linkage_num * 100 / gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                        file_out_handle.write(each_match)
                                        MarkerGene_with_assignment.add(MarkerGene)
                                    else:
                                        MarkerGene_with_assignment.add(MarkerGene)
                    # ignore depth
                    else:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            gnm_max_link_num_dict[GenomicSeq] = linkage_num
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:
                            # get identity with best marker
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                            iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                if (linkage_num*100/gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                                else:
                                    MarkerGene_with_assignment.add(MarkerGene)
    file_out_handle.close()


def get_mean_iden_list(linked_16s_list, pairwise_16s_iden_dict):

    mean_iden_list = []
    for each_16s_1 in linked_16s_list:
        current_16s_iden_list = []
        for each_16s_2 in linked_16s_list:
            if each_16s_1 != each_16s_2:
                key = '__|__'.join(sorted([each_16s_1, each_16s_2]))
                key_iden = pairwise_16s_iden_dict.get(key, 0)
                current_16s_iden_list.append(key_iden)
        mean_iden = sum(current_16s_iden_list) / len(current_16s_iden_list)
        mean_iden = float("{0:.3f}".format(mean_iden))
        mean_iden_list.append(mean_iden)

    return mean_iden_list


def filter_linkages_iteratively_new(sorted_file_in, pairwise_16s_iden_dict, within_genome_16s_divergence_cutoff, marker_len_dict,
                                    min_linkages, within_gnm_linkage_num_diff, file_out,
                                    marker_to_gnm_linking_cigar_dict_16s_side,
                                    marker_to_gnm_linking_cigar_dict_ctg_side,
                                    marker_to_ctg_gnm_Key_connector):

    # filter linkage
    gnm_max_link_num_dict = {}
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    MarkerGene_to_be_ignored = set()
    current_gnm = ''
    current_gnm_best_16s_list = []
    current_gnm_highest_link_num = 0
    gnm_with_assignment = set()
    gnm_to_assignmed_16s_dict = dict()
    best_16s_processed_gnm_list = set()
    for each_match in open(sorted_file_in):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            MarkerGene_len = marker_len_dict[MarkerGene]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])
            MarkerGene_to_GenomicSeq_key = '%s%s%s' % (MarkerGene, marker_to_ctg_gnm_Key_connector, GenomicSeq)

            if linkage_num >= min_linkages:

                # first check if linked to conserved regions
                already_assigned_16s_list = []
                iden_with_already_assigned_16s_list = []
                for already_assigned_16s in MarkerGene_with_assignment:
                    current_key = '__|__'.join(sorted([already_assigned_16s, MarkerGene]))
                    current_key_value = pairwise_16s_iden_dict.get(current_key, 0)
                    already_assigned_16s_list.append(already_assigned_16s)
                    iden_with_already_assigned_16s_list.append(current_key_value)

                if len(already_assigned_16s_list) > 0:
                    sorted_best_matched_16s_list = [[seq_id, mean_iden] for mean_iden, seq_id in sorted(zip(iden_with_already_assigned_16s_list, already_assigned_16s_list), reverse=True)]
                    best_matched_marker      = sorted_best_matched_16s_list[0][0]
                    best_matched_marker_iden = sorted_best_matched_16s_list[0][1]
                    best_matched_marker_len  = marker_len_dict[best_matched_marker]
                    if ((best_matched_marker_len - MarkerGene_len) >= 200) and (best_matched_marker_iden >= 99):

                        # get clp pct at gnm level
                        linking_cigar_16s_side = marker_to_gnm_linking_cigar_dict_16s_side[MarkerGene_to_GenomicSeq_key]
                        linking_cigar_ctg_side = marker_to_gnm_linking_cigar_dict_ctg_side[MarkerGene_to_GenomicSeq_key]
                        linking_cigar_16s_side_clp = [i for i in linking_cigar_16s_side if (('S' in i) or ('s' in i))]
                        linking_cigar_ctg_side_clp = [i for i in linking_cigar_ctg_side if (('S' in i) or ('s' in i))]
                        linking_cigar_16s_side_clp_pct = len(linking_cigar_16s_side_clp) * 100 / len(linking_cigar_16s_side)
                        linking_cigar_ctg_side_clp_pct = len(linking_cigar_ctg_side_clp) * 100 / len(linking_cigar_ctg_side)

                        if (linking_cigar_16s_side_clp_pct >= 60) and (linking_cigar_ctg_side_clp_pct >= 60):
                            MarkerGene_to_be_ignored.add(MarkerGene)

                if (MarkerGene in MarkerGene_with_assignment) or (MarkerGene in MarkerGene_to_be_ignored):
                    pass
                else:
                    if current_gnm == '':
                        current_gnm = GenomicSeq
                        current_gnm_best_16s_list.append(MarkerGene)
                        current_gnm_highest_link_num = linkage_num

                    elif current_gnm == GenomicSeq:

                        if linkage_num == current_gnm_highest_link_num:
                            current_gnm_best_16s_list.append(MarkerGene)
                        else:
                            # process markers with highest number of linking reads
                            if current_gnm not in best_16s_processed_gnm_list:

                                gnm_max_link_num_dict[current_gnm] = current_gnm_highest_link_num

                                if len(current_gnm_best_16s_list) == 1:
                                    file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (current_gnm_best_16s_list[0], current_gnm, current_gnm_highest_link_num))
                                    gnm_with_assignment.add(current_gnm)
                                    MarkerGene_with_assignment.add(current_gnm_best_16s_list[0])
                                    gnm_to_assignmed_16s_dict[current_gnm] = {current_gnm_best_16s_list[0]}
                                    best_16s_processed_gnm_list.add(current_gnm)

                                else:
                                    # process here
                                    current_gnm_best_16s_mean_iden_list = get_mean_iden_list(current_gnm_best_16s_list, pairwise_16s_iden_dict)
                                    sorted_best_16s_list = [seq_id for mean_iden, seq_id in sorted(zip(current_gnm_best_16s_mean_iden_list, current_gnm_best_16s_list), reverse=True)]

                                    add_index = 0
                                    for linked_16s in sorted_best_16s_list:
                                        if add_index == 0:
                                            file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (linked_16s, current_gnm, current_gnm_highest_link_num))
                                            gnm_with_assignment.add(current_gnm)
                                            MarkerGene_with_assignment.add(linked_16s)
                                            if current_gnm not in gnm_to_assignmed_16s_dict:
                                                gnm_to_assignmed_16s_dict[current_gnm] = {linked_16s}
                                            else:
                                                gnm_to_assignmed_16s_dict[current_gnm].add(linked_16s)
                                        else:
                                            # get identity list with already assigned 16s
                                            already_assigned_16s_list = gnm_to_assignmed_16s_dict.get(current_gnm)
                                            iden_with_already_assigned_16s_list = []
                                            for each_assigned_16s in already_assigned_16s_list:
                                                marker_key = '__|__'.join(sorted([each_assigned_16s, linked_16s]))
                                                marker_key_iden = pairwise_16s_iden_dict.get(marker_key, 0)
                                                iden_with_already_assigned_16s_list.append(marker_key_iden)
                                            if min(iden_with_already_assigned_16s_list) >= within_genome_16s_divergence_cutoff:
                                                file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (linked_16s, current_gnm, current_gnm_highest_link_num))
                                                gnm_with_assignment.add(current_gnm)
                                                MarkerGene_with_assignment.add(linked_16s)
                                                if current_gnm not in gnm_to_assignmed_16s_dict:
                                                    gnm_to_assignmed_16s_dict[current_gnm] = {linked_16s}
                                                else:
                                                    gnm_to_assignmed_16s_dict[current_gnm].add(linked_16s)
                                        add_index += 1
                                    best_16s_processed_gnm_list.add(current_gnm)

                            # process current one here
                            # get identity list with already assigned 16s
                            already_assigned_16s_list = gnm_to_assignmed_16s_dict.get(GenomicSeq)
                            iden_with_already_assigned_16s_list = []
                            for each_assigned_16s in already_assigned_16s_list:
                                marker_key = '__|__'.join(sorted([each_assigned_16s, MarkerGene]))
                                marker_key_iden = pairwise_16s_iden_dict.get(marker_key, 0)
                                iden_with_already_assigned_16s_list.append(marker_key_iden)
                            if min(iden_with_already_assigned_16s_list) >= within_genome_16s_divergence_cutoff:

                                gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                if (linkage_num * 100 / gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                    file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (MarkerGene, GenomicSeq, linkage_num))
                                    gnm_with_assignment.add(GenomicSeq)
                                    MarkerGene_with_assignment.add(MarkerGene)
                                    if GenomicSeq not in gnm_to_assignmed_16s_dict:
                                        gnm_to_assignmed_16s_dict[GenomicSeq] = {MarkerGene}
                                    else:
                                        gnm_to_assignmed_16s_dict[GenomicSeq].add(MarkerGene)
                                else:
                                    # check length, if shorter than assigned, ignore this one by adding it to MarkerGene_to_be_ignored
                                    MarkerGene_len = marker_len_dict[MarkerGene]
                                    already_assigned_16s_len_list = [marker_len_dict[s16] for s16 in already_assigned_16s_list]
                                    median_assign_16s_len = np.median(already_assigned_16s_len_list)
                                    if (median_assign_16s_len - MarkerGene_len) > 50:
                                        MarkerGene_to_be_ignored.add(MarkerGene)
                    else:
                        # first check if previous one processed
                        if current_gnm not in best_16s_processed_gnm_list:

                            gnm_max_link_num_dict[current_gnm] = current_gnm_highest_link_num

                            # process unprocessed
                            if len(current_gnm_best_16s_list) == 1:
                                file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (current_gnm_best_16s_list[0], current_gnm, current_gnm_highest_link_num))
                                gnm_with_assignment.add(current_gnm)
                                MarkerGene_with_assignment.add(current_gnm_best_16s_list[0])
                                gnm_to_assignmed_16s_dict[current_gnm] = {current_gnm_best_16s_list[0]}
                                best_16s_processed_gnm_list.add(current_gnm)
                            else:
                                current_gnm_best_16s_mean_iden_list = get_mean_iden_list(current_gnm_best_16s_list, pairwise_16s_iden_dict)
                                sorted_best_16s_list = [seq_id for mean_iden, seq_id in sorted(zip(current_gnm_best_16s_mean_iden_list, current_gnm_best_16s_list), reverse=True)]

                                add_index = 0
                                for linked_16s in sorted_best_16s_list:
                                    if add_index == 0:
                                        file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (linked_16s, current_gnm, current_gnm_highest_link_num))
                                        gnm_with_assignment.add(current_gnm)
                                        MarkerGene_with_assignment.add(linked_16s)
                                        if current_gnm not in gnm_to_assignmed_16s_dict:
                                            gnm_to_assignmed_16s_dict[current_gnm] = {linked_16s}
                                        else:
                                            gnm_to_assignmed_16s_dict[current_gnm].add(linked_16s)
                                    else:
                                        # get identity list with already assigned 16s
                                        already_assigned_16s_list = gnm_to_assignmed_16s_dict.get(current_gnm)
                                        iden_with_already_assigned_16s_list = []
                                        for each_assigned_16s in already_assigned_16s_list:
                                            marker_key = '__|__'.join(sorted([each_assigned_16s, linked_16s]))
                                            marker_key_iden = pairwise_16s_iden_dict.get(marker_key, 0)
                                            iden_with_already_assigned_16s_list.append(marker_key_iden)

                                        if min(iden_with_already_assigned_16s_list) >= within_genome_16s_divergence_cutoff:
                                            file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (linked_16s, current_gnm, current_gnm_highest_link_num))
                                            gnm_with_assignment.add(current_gnm)
                                            MarkerGene_with_assignment.add(linked_16s)
                                            if current_gnm not in gnm_to_assignmed_16s_dict:
                                                gnm_to_assignmed_16s_dict[current_gnm] = {linked_16s}
                                            else:
                                                gnm_to_assignmed_16s_dict[current_gnm].add(linked_16s)
                                    add_index += 1
                                best_16s_processed_gnm_list.add(current_gnm)

                        # then process current one
                        if GenomicSeq in best_16s_processed_gnm_list:
                            # get identity list with already assigned 16s
                            already_assigned_16s_list = gnm_to_assignmed_16s_dict.get(GenomicSeq)


                            iden_with_already_assigned_16s_list = []
                            for each_assigned_16s in already_assigned_16s_list:
                                marker_key = '__|__'.join(sorted([each_assigned_16s, MarkerGene]))
                                marker_key_iden = pairwise_16s_iden_dict.get(marker_key, 0)
                                iden_with_already_assigned_16s_list.append(marker_key_iden)

                            if min(iden_with_already_assigned_16s_list) >= within_genome_16s_divergence_cutoff:

                                gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                if (linkage_num * 100 / gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                    file_out_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (MarkerGene, GenomicSeq, linkage_num))
                                    gnm_with_assignment.add(GenomicSeq)
                                    MarkerGene_with_assignment.add(MarkerGene)
                                    if GenomicSeq not in gnm_to_assignmed_16s_dict:
                                        gnm_to_assignmed_16s_dict[GenomicSeq] = {MarkerGene}
                                    else:
                                        gnm_to_assignmed_16s_dict[GenomicSeq].add(MarkerGene)
                                else:
                                    # check length, if shorter than assigned, ignore this one by adding it to MarkerGene_with_assignment
                                    already_assigned_16s_len_list = [marker_len_dict[s16] for s16 in already_assigned_16s_list]
                                    median_assign_16s_len = np.median(already_assigned_16s_len_list)
                                    if (median_assign_16s_len - MarkerGene_len) > 50:
                                        MarkerGene_to_be_ignored.add(MarkerGene)
                        else:
                            current_gnm = GenomicSeq
                            current_gnm_best_16s_list = [MarkerGene]
                            current_gnm_highest_link_num = linkage_num

    #print('current_gnm: %s' % current_gnm)
    #print('current_gnm_highest_link_num: %s' % current_gnm_highest_link_num)
    #print('current_gnm_best_16s_list: %s' % current_gnm_best_16s_list)

    file_out_handle.close()


def combine_paired_and_clipping_linkages(paired_linkages, clipping_linkages, file_out_summary, file_out_intersect_linkages):

    # file in:   file_in_paired    and  file_in_clipping
    # file out:  file_out_summary  and  file_out_intersection

    combined_paired_and_clipping_keys = set()

    # read in paired linkages
    paired_linkages_dict = {}
    for paired_linkage in open(paired_linkages):
        if not paired_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            paired_linkage_split = paired_linkage.strip().split(',')
            paired_key = '%s__|__%s' % (paired_linkage_split[0], paired_linkage_split[1])
            paired_value = int(paired_linkage_split[2])
            paired_linkages_dict[paired_key] = paired_value
            combined_paired_and_clipping_keys.add(paired_key)

    # read in clipping linkages
    clipping_linkages_dict = {}
    for clipping_linkage in open(clipping_linkages):
        if not clipping_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            clipping_linkage_split = clipping_linkage.strip().split(',')
            clipping_key = '%s__|__%s' % (clipping_linkage_split[0], clipping_linkage_split[1])
            clipping_value = int(clipping_linkage_split[2])
            clipping_linkages_dict[clipping_key] = clipping_value
            combined_paired_and_clipping_keys.add(clipping_key)

    combined_paired_and_clipping_keys_sorted = sorted([i for i in combined_paired_and_clipping_keys])

    # combine paired and clipping linkages
    file_out_summary_handle = open(file_out_summary, 'w')
    file_out_intersect_linkages_handle = open(file_out_intersect_linkages, 'w')
    file_out_summary_handle.write('MarkerGene\tGenomicSeq\tPaired\tClipping\n')
    file_out_intersect_linkages_handle.write('MarkerGene,GenomicSeq,Number\n')
    for each_key in combined_paired_and_clipping_keys_sorted:

        current_key_paired_value = 0
        if each_key in paired_linkages_dict:
            current_key_paired_value = paired_linkages_dict[each_key]

        current_key_clipping_value = 0
        if each_key in clipping_linkages_dict:
            current_key_clipping_value = clipping_linkages_dict[each_key]

        if current_key_paired_value > 0:

            current_key_combined = current_key_paired_value + current_key_clipping_value

            # write out
            file_out_summary_handle.write('%s\t%s\t%s\n' % ('\t'.join([i[12:] for i in each_key.split('__|__')]), current_key_paired_value, current_key_clipping_value))
            file_out_intersect_linkages_handle.write('%s,%s\n' % (','.join(each_key.split('__|__')), current_key_combined))

    file_out_summary_handle.close()
    file_out_intersect_linkages_handle.close()


def get_unlinked_mag_end_seq(ref_in, ref_in_end_seq, end_seq_len, ctg_ignore_region_dict_rd1):
    ctg_ignore_region_dict_rd2 = dict()

    # get ref seqs subset
    ref_subset_handle = open(ref_in_end_seq, 'w')
    for ref_seq in SeqIO.parse(ref_in, 'fasta'):

        ref_seq_id = ref_seq.id
        ref_seq_len = len(ref_seq.seq)

        if ref_seq_len < end_seq_len * 2:
            ref_subset_handle.write('>%s\n' % ref_seq_id)
            ref_subset_handle.write('%s\n' % ref_seq.seq)

            # add to ctg_ignore_region_dict_rd2
            if ref_seq_id in ctg_ignore_region_dict_rd1:
                ctg_ignore_region_dict_rd2[ref_seq_id] = ctg_ignore_region_dict_rd1[ref_seq_id]
        else:
            ref_seq_left_end_id = '%s_l' % ref_seq_id
            ref_seq_right_end_id = '%s_r' % ref_seq_id
            ref_seq_left_end = ref_seq.seq[:end_seq_len]
            ref_seq_right_end = ref_seq.seq[-end_seq_len:]

            # write out left end
            ref_subset_handle.write('>%s\n' % ref_seq_left_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_left_end)

            # write out right end
            ref_subset_handle.write('>%s\n' % ref_seq_right_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_right_end)

            # add to ctg_ignore_region_dict_rd2
            if ref_seq_id in ctg_ignore_region_dict_rd1:
                current_seq_to_ignore_ends = ctg_ignore_region_dict_rd1[ref_seq_id]
                for end_to_ignore in current_seq_to_ignore_ends:
                    if end_to_ignore == 'left_end':
                        ctg_ignore_region_dict_rd2[ref_seq_left_end_id] = {'left_end'}
                    if end_to_ignore == 'right_end':
                        ctg_ignore_region_dict_rd2[ref_seq_right_end_id] = {'right_end'}
    ref_subset_handle.close()

    return ctg_ignore_region_dict_rd2


def get_free_living_mate(ref_in, sam_file, reads_r1, reads_r2, end_seq_len, num_threads, pwd_bbmap_exe, bbmap_memory):

    ref_in_path, ref_in_basename, ref_in_ext = sep_path_basename_ext(ref_in)

    ref_subset      = '%s/%s_ends_%sbp%s'                % (ref_in_path, ref_in_basename, end_seq_len, ref_in_ext)
    bbmap_stderr    = '%s/%s_ends_%sbp_bbmap_stderr.txt' % (ref_in_path, ref_in_basename, end_seq_len)

    # get ref seqs subset
    ref_subset_handle = open(ref_subset, 'w')
    for ref_seq in SeqIO.parse(ref_in, 'fasta'):

        ref_seq_id = ref_seq.id
        ref_seq_len = len(ref_seq.seq)

        if ref_seq_len < end_seq_len * 2:
            ref_subset_handle.write('>%s\n' % ref_seq_id)
            ref_subset_handle.write('%s\n' % ref_seq.seq)
        else:
            ref_seq_left_end_id = '%s_l' % ref_seq_id
            ref_seq_right_end_id = '%s_r' % ref_seq_id
            ref_seq_left_end = ref_seq.seq[:end_seq_len]
            ref_seq_right_end = ref_seq.seq[-end_seq_len:]

            # write out left end
            ref_subset_handle.write('>%s\n' % ref_seq_left_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_left_end)

            # write out right end
            ref_subset_handle.write('>%s\n' % ref_seq_right_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_right_end)
    ref_subset_handle.close()

    # mapping with bbmap
    bbmap_parameter_round2 = 'local=t nodisk=t ambiguous=all keepnames=t saa=f trd=t silent=true threads=%s -Xmx%sg' % (num_threads, bbmap_memory)
    bbmap_cmd_round2 = '%s ref=%s in=%s in2=%s outm=%s %s 2> %s' % (pwd_bbmap_exe, ref_subset, reads_r1, reads_r2, sam_file, bbmap_parameter_round2, bbmap_stderr)
    os.system(bbmap_cmd_round2)

    # mapping with bowtie


def get_best_ctg_or_16s_for_gap_seq_iteratively(file_in, sort_by_col_header, min_linkages, file_out):

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    file_out_handle = open(file_out, 'w')
    gap_seq_with_assignment = set()
    for each_match in open(file_in_sorted):
        if each_match.startswith('Gap_seq,'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            gap_seq_id = match_split[0]
            linkage_num = int(match_split[2])
            if (linkage_num >= min_linkages) and (gap_seq_id not in gap_seq_with_assignment):
                file_out_handle.write(each_match)
                gap_seq_with_assignment.add(gap_seq_id)
    file_out_handle.close()

    # remove tmp file
    # os.remove(file_in_sorted)


def get_accuracy(file_in, marker_num):

    linkage_num_total = 0
    linkage_num_correct = 0
    recovered_markers = set()
    for each_match in open(file_in):
        if not each_match.startswith('MarkerGene\tGenomicSeq\tLinkage'):
            match_split = each_match.strip().split('\t')
            linkage_num = int(match_split[2])
            MarkerGene_genome = match_split[0][:2]
            GenomicSeq_genome = match_split[1]

            linkage_num_total += linkage_num
            if MarkerGene_genome == GenomicSeq_genome:
                linkage_num_correct += linkage_num
                recovered_markers.add(match_split[0])

    marker_recovery = float("{0:.2f}".format(len(recovered_markers)*100/marker_num))

    link_accuracy = 0
    if linkage_num_total > 0:
        link_accuracy = float("{0:.2f}".format(linkage_num_correct*100/linkage_num_total))

    marker_recovery = '%s/%s(%s)' % (len(recovered_markers), marker_num, marker_recovery)

    return marker_recovery, link_accuracy, recovered_markers


def get_accuracy_by_genome(file_in, mag_folder, mag_file_extension):

    # get MAG file list
    mag_file_re             = '%s/*%s' % (mag_folder, mag_file_extension)
    mag_file_list           = [os.path.basename(file_name) for file_name in glob.glob(mag_file_re)]
    mag_file_list_no_ext    = {'.'.join(i.split('.')[:-1]) for i in mag_file_list}

    genome_with_right_16s_assignment_tmp = set()
    genome_with_wrong_16s_assignment = set()
    for each_match in open(file_in):
        if not each_match.startswith('MarkerGene\tGenomicSeq\tLinkage'):
            match_split = each_match.strip().split('\t')
            MarkerGene_genome = match_split[0][:2]
            GenomicSeq_genome = match_split[1]

            if GenomicSeq_genome == MarkerGene_genome:
                genome_with_right_16s_assignment_tmp.add(GenomicSeq_genome)
            else:
                genome_with_wrong_16s_assignment.add(GenomicSeq_genome)

    genome_with_right_16s_assignment_always = []
    genome_without_right_16s_assignment = []
    for input_genome in mag_file_list_no_ext:
        if (input_genome in genome_with_right_16s_assignment_tmp) and (input_genome not in genome_with_wrong_16s_assignment):
            genome_with_right_16s_assignment_always.append(input_genome)
        else:
            genome_without_right_16s_assignment.append(input_genome)


    marker_gene_assignment_rate = float("{0:.2f}".format(len(genome_with_right_16s_assignment_always)*100/len(mag_file_list_no_ext)))

    marker_gene_assignment_accuracy = 0
    if (len(genome_with_right_16s_assignment_always) + len(genome_with_wrong_16s_assignment)) > 0:
        marker_gene_assignment_accuracy = float("{0:.2f}".format(len(genome_with_right_16s_assignment_always)*100/(len(genome_with_right_16s_assignment_always) + len(genome_with_wrong_16s_assignment))))
    marker_gene_assignment_rate = '%s/%s(%s)' % (len(genome_with_right_16s_assignment_always), len(mag_file_list_no_ext), marker_gene_assignment_rate)

    return marker_gene_assignment_rate, marker_gene_assignment_accuracy, genome_with_right_16s_assignment_always, genome_without_right_16s_assignment


def rename_seq(ctg_file_in, ctg_file_out, prefix, str_connector):

    ctg_file_out_handle = open(ctg_file_out, 'w')
    for Seq_record in SeqIO.parse(ctg_file_in, 'fasta'):
        Seq_record.id = '%s%s%s' % (prefix, str_connector, Seq_record.id)
        SeqIO.write(Seq_record, ctg_file_out_handle, 'fasta')
    ctg_file_out_handle.close()


def SeqIO_convert_worker(argument_list):

    file_in         = argument_list[0]
    file_in_fmt     = argument_list[1]
    file_out        = argument_list[2]
    file_out_fmt    = argument_list[3]
    SeqIO.convert(file_in, file_in_fmt, file_out, file_out_fmt)


def get_max_clp_and_index(r1_cigar_list, r2_cigar_list):

    r1_cigar_list_split = [cigar_splitter(i) for i in r1_cigar_list]
    r2_cigar_list_split = [cigar_splitter(i) for i in r2_cigar_list]

    r1_cigar_list_split_only_clp = []
    for each_r1_cigar_split in r1_cigar_list_split:
        clp_len_l = 0
        if each_r1_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r1_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r1_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r1_cigar_split[-1][:-1])
        r1_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    r2_cigar_list_split_only_clp = []
    for each_r2_cigar_split in r2_cigar_list_split:
        clp_len_l = 0
        if each_r2_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r2_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r2_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r2_cigar_split[-1][:-1])
        r2_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    cigar_list_split_only_clp_r1_r2 = [r1_cigar_list_split_only_clp, r2_cigar_list_split_only_clp]

    max_value = 0
    max_value_index = ''
    for num_list_1 in cigar_list_split_only_clp_r1_r2[0]:
        if num_list_1[0] > max_value:
            max_value = num_list_1[0]
            max_value_index = 'r1_l'
        if num_list_1[1] > max_value:
            max_value = num_list_1[1]
            max_value_index = 'r1_r'
    for num_list_2 in cigar_list_split_only_clp_r1_r2[1]:
        if num_list_2[0] > max_value:
            max_value = num_list_2[0]
            max_value_index = 'r2_l'
        if num_list_2[1] > max_value:
            max_value = num_list_2[1]
            max_value_index = 'r2_r'

    # get the best cigar
    best_cigar = ''
    if max_value_index == 'r1_l':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r1_r':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_l':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_r':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar

    return best_cigar, max_value, max_value_index


class MappingRecord:

    def __init__(self):

        #################### overall ####################

        self.qualified_reads = False

        #################### round 1 16s ####################

        self.consider_r1_unmapped_mate = False
        self.consider_r2_unmapped_mate = False

        self.r1_16s_ref_dict = dict()
        self.r2_16s_ref_dict = dict()

        self.r1_16s_refs_lowest_mismatch = None
        self.r2_16s_refs_lowest_mismatch = None

        self.r1_16s_refs_no_ignored = dict()
        self.r2_16s_refs_no_ignored = dict()
        self.shared_16s_refs_no_ignored = dict()

        self.both_mapped_to_16s = False

        #################### round 1 ctg ####################

        self.r1_ctg_ref_dict = dict()
        self.r2_ctg_ref_dict = dict()

        self.r1_ctg_refs_lowest_mismatch = None
        self.r2_ctg_refs_lowest_mismatch = None

        self.r1_ctg_refs_no_ignored = dict()
        self.r2_ctg_refs_no_ignored = dict()
        self.shared_ctg_refs_no_ignored = dict()

        self.matched_to_ctg = False

        #################### round 2 ####################

        self.qualified_reads_rd2 = False

        self.r1_ctg_ref_dict_rd2 = dict()
        self.r2_ctg_ref_dict_rd2 = dict()

        self.r1_ctg_refs_lowest_mismatch_rd2 = None
        self.r2_ctg_refs_lowest_mismatch_rd2 = None

        #################### round 2 mini_assembly ####################

        self.r1_mini_ref_dict = dict()
        self.r2_mini_ref_dict = dict()

        self.r1_mini_refs_lowest_mismatch = None
        self.r2_mini_refs_lowest_mismatch = None

        self.r1_mini_refs_no_ignored = dict()
        self.r2_mini_refs_no_ignored = dict()
        self.shared_mini_refs_no_ignored = dict()


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len):

    mismatch_set_all_cigar = set()
    mismatch_set_long_M_cigars = set()
    for each_cigar in r1_ref_cigar_set:
        aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
        mismatch_set_all_cigar.add(mismatch_pct)
        if aligned_len >= min_M_len:
            mismatch_set_long_M_cigars.add(mismatch_pct)

    min_mismatch = 'NA'
    if len(mismatch_set_all_cigar) > 0:
        min_mismatch = min(mismatch_set_all_cigar)
        if len(mismatch_set_long_M_cigars) > 0:
            min_mismatch = min(mismatch_set_long_M_cigars)

    return min_mismatch


def get_sankey_plot(node_list, source_list, target_list, value_list, color_list, plot_title, plot_height, output_html):

    node_index_dict = {y: x for x, y in enumerate(node_list)}
    source_index = [node_index_dict[x] for x in source_list]
    target_index = [node_index_dict[x] for x in target_list]

    # https://anvil.works/docs/api/plotly.graph_objs.sankey
    fig = go.Figure(data=[go.Sankey(node=dict(label=node_list,  # line=0,
                                              pad=5,  # space between node
                                              thickness=12,  # node width
                                              line=dict(width=0)),  # set width of node border to 0
                                    link=dict(source=source_index,
                                              target=target_index,
                                              value=value_list,
                                              color=color_list))])

    fig.update_layout(autosize=False, width=1200, height=plot_height, margin=dict(l=50, r=50, b=50, t=125), paper_bgcolor="white", title=plot_title)
    fig.update_traces(textfont_size=11)
    fig.write_html(output_html)


def sankey_linkages(combined_linkage_file_ctg_level, linkage_plot_rd1_html, linkage_plot_rd2_html):

    dict_for_sankey_key_connector   = '___X___'

    linkage_num_dict_rd1 = {}
    linkage_num_dict_rd2 = {}
    ctg_to_gnm_dict_rd1 = {}
    ctg_to_gnm_dict_rd2 = {}
    node_set_rd1 = set()
    node_set_rd2 = set()
    genome_set_rd1 = set()
    genome_set_rd2 = set()
    contig_set_rd1 = set()
    contig_set_rd2 = set()
    marker_gene_set_rd1 = set()
    marker_gene_set_rd2 = set()
    for each_linkage in open(combined_linkage_file_ctg_level):
        if not each_linkage.startswith('Marker___Genome(total)\tContig\tRd1\tRd2'):
            each_linkage_split = each_linkage.strip().split('\t')

            marker_id = each_linkage_split[0].split('___')[0]
            gnm_id = each_linkage_split[0].split('___')[1].split('(')[0]
            ctg_id = each_linkage_split[1]
            total_link_num = int(each_linkage_split[2]) + int(each_linkage_split[3])
            marker_to_ctg_key = '%s%s%s' % (marker_id, dict_for_sankey_key_connector, ctg_id)
            ctg_to_gnm_key = '%s%s%s' % (ctg_id, dict_for_sankey_key_connector, gnm_id)

            if int(each_linkage_split[3]) == 0:
                genome_set_rd1.add(gnm_id)
                contig_set_rd1.add(ctg_id)
                marker_gene_set_rd1.add(marker_id)
                node_set_rd1.add(marker_id)
                node_set_rd1.add(ctg_id)
                node_set_rd1.add(gnm_id)

                if ctg_id not in ctg_to_gnm_dict_rd1:
                    ctg_to_gnm_dict_rd1[ctg_id] = gnm_id

                if marker_to_ctg_key not in linkage_num_dict_rd1:
                    linkage_num_dict_rd1[marker_to_ctg_key] = total_link_num
                else:
                    linkage_num_dict_rd1[marker_to_ctg_key] += total_link_num

                if ctg_to_gnm_key not in linkage_num_dict_rd1:
                    linkage_num_dict_rd1[ctg_to_gnm_key] = total_link_num
                else:
                    linkage_num_dict_rd1[ctg_to_gnm_key] += total_link_num

            if int(each_linkage_split[2]) == 0:
                genome_set_rd2.add(gnm_id)
                contig_set_rd2.add(ctg_id)
                marker_gene_set_rd2.add(marker_id)
                node_set_rd2.add(marker_id)
                node_set_rd2.add(ctg_id)
                node_set_rd2.add(gnm_id)
                if ctg_id not in ctg_to_gnm_dict_rd2:
                    ctg_to_gnm_dict_rd2[ctg_id] = gnm_id

                if marker_to_ctg_key not in linkage_num_dict_rd2:
                    linkage_num_dict_rd2[marker_to_ctg_key] = total_link_num
                else:
                    linkage_num_dict_rd2[marker_to_ctg_key] += total_link_num

                if ctg_to_gnm_key not in linkage_num_dict_rd2:
                    linkage_num_dict_rd2[ctg_to_gnm_key] = total_link_num
                else:
                    linkage_num_dict_rd2[ctg_to_gnm_key] += total_link_num

    source_list_rd1 = []
    target_list_rd1 = []
    value_list_rd1 = []
    for each_rd1_linkage in linkage_num_dict_rd1:
        each_rd1_linkage_split = each_rd1_linkage.split(dict_for_sankey_key_connector)
        source_list_rd1.append(each_rd1_linkage_split[0])
        target_list_rd1.append(each_rd1_linkage_split[1])
        value_list_rd1.append(linkage_num_dict_rd1[each_rd1_linkage])

    source_list_rd2 = []
    target_list_rd2 = []
    value_list_rd2 = []
    for each_rd2_linkage in linkage_num_dict_rd2:
        each_rd2_linkage_split = each_rd2_linkage.split(dict_for_sankey_key_connector)
        source_list_rd2.append(each_rd2_linkage_split[0])
        target_list_rd2.append(each_rd2_linkage_split[1])
        value_list_rd2.append(linkage_num_dict_rd2[each_rd2_linkage])

    gnm_color_list_rd1 = sns.color_palette('tab20', len(genome_set_rd1)).as_hex()
    gnm_color_list_rd2 = sns.color_palette('tab20', len(genome_set_rd2)).as_hex()

    genome_to_color_dict_rd1 = {gnm: color for gnm, color in zip(genome_set_rd1, gnm_color_list_rd1)}
    genome_to_color_dict_rd2 = {gnm: color for gnm, color in zip(genome_set_rd2, gnm_color_list_rd2)}

    color_list_rd1 = []
    for each_target in target_list_rd1:
        if each_target in genome_to_color_dict_rd1:
            color_list_rd1.append(genome_to_color_dict_rd1[each_target])
        else:
            target_genome = ctg_to_gnm_dict_rd1[each_target]
            color_list_rd1.append(genome_to_color_dict_rd1[target_genome])

    color_list_rd2 = []
    for each_target in target_list_rd2:
        if each_target in genome_to_color_dict_rd2:
            color_list_rd2.append(genome_to_color_dict_rd2[each_target])
        else:
            target_genome = ctg_to_gnm_dict_rd2[each_target]
            color_list_rd2.append(genome_to_color_dict_rd2[target_genome])

    node_list_rd1 = sorted([i for i in node_set_rd1])
    node_list_rd2 = sorted([i for i in node_set_rd2])

    plot_title_text_rd1 = 'MarkerMAG detected linkages (round 1)<br>Number of linked genomes: %s<br>Number of linked markers: %s' % (len(genome_set_rd1), len(marker_gene_set_rd1))
    plot_title_text_rd2 = 'MarkerMAG detected linkages (round 2)<br>Number of linked genomes: %s<br>Number of linked markers: %s' % (len(genome_set_rd2), len(marker_gene_set_rd2))

    plot_height_rd1 = 900 if max([len(contig_set_rd1), len(marker_gene_set_rd1)]) <= 25 else max([len(contig_set_rd1), len(marker_gene_set_rd1)]) * 32
    plot_height_rd2 = 900 if max([len(contig_set_rd2), len(marker_gene_set_rd2)]) <= 25 else max([len(contig_set_rd2), len(marker_gene_set_rd2)]) * 32

    plot_title_dict_rd1 = dict(text=plot_title_text_rd1, x=0.05, y=(1-(50/plot_height_rd1)))
    plot_title_dict_rd2 = dict(text=plot_title_text_rd2, x=0.05, y=(1-(50/plot_height_rd2)))

    get_sankey_plot(node_list_rd1, source_list_rd1, target_list_rd1, value_list_rd1, color_list_rd1, plot_title_dict_rd1, plot_height_rd1, linkage_plot_rd1_html)
    get_sankey_plot(node_list_rd2, source_list_rd2, target_list_rd2, value_list_rd2, color_list_rd2, plot_title_dict_rd2, plot_height_rd2, linkage_plot_rd2_html)


def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


def remove_high_mismatch(sam_in, mismatch_cutoff, sam_out):

    sam_out_handle = open(sam_out, 'w')
    ref_len_dict = {}
    for each_read in open(sam_in):
        each_read_split = each_read.strip().split('\t')
        if each_read.startswith('@'):
            sam_out_handle.write(each_read)

            marker_id = ''
            marker_len = 0
            for each_element in each_read_split:
                if each_element.startswith('SN:'):
                    marker_id = each_element[3:]
                if each_element.startswith('LN:'):
                    marker_len = int(each_element[3:])
            ref_len_dict[marker_id] = marker_len

        else:
            ref_id = each_read_split[2]
            ref_pos = int(each_read_split[3])
            cigar = each_read_split[5]
            if cigar == '*':
                sam_out_handle.write(each_read)
            else:
                cigar_splitted = cigar_splitter(cigar)
                both_ends_clp = check_both_ends_clipping(cigar_splitted)
                if both_ends_clp is False:
                    r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(cigar_splitted)
                    if r1_mismatch_pct <= mismatch_cutoff:

                        # check if clp in middle
                        if ('S' not in cigar) and ('s' not in cigar):
                            sam_out_handle.write(each_read)
                        else:
                            clip_in_middle = True
                            if (cigar_splitted[0][-1] in ['S', 's']) and (ref_pos == 1):
                                clip_in_middle = False
                            if (cigar_splitted[-1][-1] in ['S', 's']):
                                if (ref_pos + r1_aligned_len - 1) == ref_len_dict[ref_id]:
                                    clip_in_middle = False

                            if clip_in_middle is False:
                                sam_out_handle.write(each_read)
    sam_out_handle.close()


def reads2bam(args):

    output_prefix   = args['p']
    ref_seq         = args['ref']
    index_ref       = args['index_ref']
    r1_seq          = args['r1']
    r2_seq          = args['r2']
    unpaired_seq    = args['u']
    fq_format       = args['fastq']
    local_aln       = args['local']
    max_mismatch    = args['mismatch']
    thread_num      = args['t']
    keep_tmp        = args['tmp']

    ref_path, ref_basename, ref_ext = sep_path_basename_ext(ref_seq)

    cmd_bowtie2_build   = 'bowtie2-build -f %s %s --threads %s' % (ref_seq, ref_basename, thread_num)

    bowtie2_parameter = '--no-unal --xeq'
    if local_aln is True:
        bowtie2_parameter = '--local --no-unal --xeq'

    cmd_bowtie2 = ''
    if (r1_seq is not None) and (r2_seq is not None) and (unpaired_seq is None):
        cmd_bowtie2     = 'bowtie2 -x %s -1 %s -2 %s -S %s_raw.sam -p %s -f %s' % (ref_basename, r1_seq, r2_seq, output_prefix, thread_num, bowtie2_parameter)
        if fq_format is True:
            cmd_bowtie2 = 'bowtie2 -x %s -1 %s -2 %s -S %s_raw.sam -p %s -q %s' % (ref_basename, r1_seq, r2_seq, output_prefix, thread_num, bowtie2_parameter)

    elif (r1_seq is not None) and (r2_seq is not None) and (unpaired_seq is not None):
        cmd_bowtie2     = 'bowtie2 -x %s -1 %s -2 %s -U %s -S %s_raw.sam -p %s -f %s' % (ref_basename, r1_seq, r2_seq, unpaired_seq, output_prefix, thread_num, bowtie2_parameter)
        if fq_format is True:
            cmd_bowtie2 = 'bowtie2 -x %s -1 %s -2 %s -U %s -S %s_raw.sam -p %s -q %s' % (ref_basename, r1_seq, r2_seq, unpaired_seq, output_prefix, thread_num, bowtie2_parameter)

    elif (r1_seq is None) and (r2_seq is None) and (unpaired_seq is not None):
        cmd_bowtie2     = 'bowtie2 -x %s -U %s -S %s_raw.sam -p %s -f %s' % (ref_basename, unpaired_seq, output_prefix, thread_num, bowtie2_parameter)
        if fq_format is True:
            cmd_bowtie2 = 'bowtie2 -x %s -U %s -S %s_raw.sam -p %s -q %s' % (ref_basename, unpaired_seq, output_prefix, thread_num, bowtie2_parameter)
    else:
        print('Please check your input reads files')
        exit()

    cmd_samtools_view   = 'samtools view -bS %s.sam -o %s.bam' % (output_prefix, output_prefix)
    cmd_samtools_sort   = 'samtools sort %s.bam -o %s_sorted.bam' % (output_prefix, output_prefix)
    cmd_samtools_index  = 'samtools index %s_sorted.bam' % output_prefix

    if index_ref is True:
        os.system(cmd_bowtie2_build)
    os.system(cmd_bowtie2)


    if max_mismatch == 100:
        os.system('mv %s_raw.sam %s.sam' % (output_prefix, output_prefix))
    else:
        remove_high_mismatch(('%s_raw.sam' % output_prefix), max_mismatch, ('%s.sam' % output_prefix))
        os.system('rm %s_raw.sam' % output_prefix)


    os.system(cmd_samtools_view)
    os.system(cmd_samtools_sort)
    os.system(cmd_samtools_index)

    if keep_tmp is False:
        os.system('rm %s.sam' % output_prefix)
        os.system('rm %s.bam' % output_prefix)


if __name__ == '__main__':

    reads2bam_parser = argparse.ArgumentParser(usage=reads2bam_usage)

    reads2bam_parser.add_argument('-p',               required=True,                                     help='output prefix')
    reads2bam_parser.add_argument('-ref',             required=True,                                     help='reference sequences')
    reads2bam_parser.add_argument('-index_ref',       required=False, action="store_true",               help='index reference')
    reads2bam_parser.add_argument('-r1',              required=False, default=None,                      help='paired reads r1')
    reads2bam_parser.add_argument('-r2',              required=False, default=None,                      help='paired reads r2')
    reads2bam_parser.add_argument('-u',               required=False, default=None,                      help='unpaired reads')
    reads2bam_parser.add_argument('-fastq',           required=False, action="store_true",               help='reads in fastq format')
    reads2bam_parser.add_argument('-local',           required=False, action="store_true",               help='perform local alignment')
    reads2bam_parser.add_argument('-mismatch',        required=False, type=int, default=100,             help='maximum mismatch pct allowed, between 1-100, default: 100')
    reads2bam_parser.add_argument('-t',               required=False, type=int, default=1,               help='number of threads, default: 1')
    reads2bam_parser.add_argument('-tmp',             required=False, action="store_true",               help='keep temporary files')

    args = vars(reads2bam_parser.parse_args())

    reads2bam(args)



import os
import shutil
import argparse
import configparser

usage = """

    Usage:

    python3 path/to/this/script.py

    It will:
    1. Combine all gbk files (from Prokka) in defined folder
    2. Combine all ffn files (from Prokka) in defined folder
    3. Make blastdb with combined ffn file
    4. Generate qsub file to run all versus all blast of combined ffn
    5. Submit generated qsub file

"""

####################################################### Configuration ##################################################

path_to_makeblastdb_executable = 'makeblastdb'
blast_module = 'blast+/2.6.0'
prokka_output = 'prokka_output'
########################################################################################################################


# get combined gbk file
os.system('cat ./%s/*/*.gbk > combined.gbk' % prokka_output)

# get combined ffn file
os.system('cat ./%s/*/*.ffn > combined.ffn' % prokka_output)

# create blastdb folder and run makeblastdb
if not os.path.isdir('qsub_files'):
    os.mkdir('./qsub_files')

if os.path.isdir('blastdb'):
    shutil.rmtree('blastdb')
os.mkdir('blastdb')

os.system('cp combined.ffn ./blastdb/')
makeblastdb_cmd = path_to_makeblastdb_executable + ' -in ./blastdb/combined.ffn -dbtype nucl -parse_seqids'
os.system(makeblastdb_cmd)

# prepare blastn qsub file header
line_1 = '#!/bin/bash\n'
line_2 = '#PBS -l nodes=1:ppn=3\n'
line_3 = '#PBS -l vmem=20gb\n'
line_4 = '#PBS -l walltime=11:59:00' + '\n'
line_5 = '#PBS -j oe\n'
line_6 = '#PBS -M wythe1987@163.com\n'
line_7 = '#PBS -m ae\n'
line_8 = 'cd ' + os.getcwd()
header = line_1 + line_2 + line_3 + line_4 + line_5 + line_6 + line_7 + line_8

qsub_blastn_file_name = 'qsub_blastn_combined_ffn.sh'
blast_qsub_out = open('./qsub_files/' + qsub_blastn_file_name, 'w')
blast_qsub_out.write(header + '\n\n')
blast_qsub_out.write('module load %s\n' % blast_module)
blast_parameters = '-evalue 1e-5 -outfmt "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen" -task blastn'
blast_qsub_out.write('blastn -query %s/combined.ffn -db %s/blastdb/combined.ffn -out %s/all_vs_all_ffn.tab %s' % (os.getcwd(), os.getcwd(), os.getcwd(), blast_parameters))
blast_qsub_out.close()

#submit qsub files
os.system('qsub ./qsub_files/%s' % qsub_blastn_file_name)



#!/usr/bin/env python
import os
from Bio import SeqIO
from Bio.Graphics import GenomeDiagram
from Bio.Graphics.GenomeDiagram import CrossLink
from reportlab.lib import colors
from reportlab.lib.units import cm


def set_contig_track_features(gene_contig, name_group_dict, candidate_list, feature_set):
    # add features to feature set
    for feature in gene_contig.features:
        if feature.type == "CDS":
            # define label color
            if feature.qualifiers['locus_tag'][0] in candidate_list:
                label_color = colors.blue
            else:
                label_color = colors.black

            # change gene name
            bin_name_gbk_split = feature.qualifiers['locus_tag'][0].split('_')
            bin_name_gbk = '_'.join(bin_name_gbk_split[:-1])
            feature.qualifiers['locus_tag'][0] = name_group_dict[bin_name_gbk] + '_' + bin_name_gbk_split[-1]

            # strands
            color = None
            label_angle = 0
            if feature.location.strand == 1:
                label_angle = 45
                color = colors.lightblue
            elif feature.location.strand == -1:
                label_angle = -225
                color = colors.lightgreen
            # add feature
            feature_set.add_feature(feature,
                                    color = color,
                                    label = True,
                                    sigil = 'ARROW',
                                    arrowshaft_height = 0.7,
                                    arrowhead_length = 0.4,
                                    label_color = label_color,
                                    label_size = 10,
                                    label_angle = label_angle,
                                    label_position = "middle")


file_1 = '16s.fasta'
file_2 = 'ctg.fasta'
path_to_blast_result = 'blast_results.txt'
plot_file_name = 'test'

# Run Blast
parameters_c_n = '-evalue 1e-5 -outfmt 6 -task blastn'
command_blast = 'blastn -query %s -subject %s -out %s %s' % (file_1, file_2, path_to_blast_result, parameters_c_n)
os.system(command_blast)

# read in fasta files
matche_pair_list = []
for each_file in [file_1, file_2]:
    gene_contig = SeqIO.read(each_file, "fasta")
    matche_pair_list.append(gene_contig)
bin_record_list = []
bin_record_list.append(matche_pair_list)

# create an empty diagram
diagram = GenomeDiagram.Diagram()

# add tracks to diagram
max_len = 0
for gene1_contig, gene2_contig in bin_record_list:
    # set diagram track length
    max_len = max(max_len, len(gene1_contig), len(gene2_contig))

    # add gene content track for gene1_contig
    contig_1_gene_content_track = diagram.new_track(1,
                                                    name=gene1_contig.name,
                                                    greytrack=True,
                                                    greytrack_labels=1,
                                                    greytrack_font='Helvetica',
                                                    greytrack_fontsize=12,
                                                    height=0.35,
                                                    start=0,
                                                    end=len(gene1_contig),
                                                    scale=True,
                                                    scale_fontsize=6,
                                                    scale_ticks=1,
                                                    scale_smalltick_interval=10000,
                                                    scale_largetick_interval=10000)

    # add gene content track for gene2_contig
    contig_2_gene_content_track = diagram.new_track(1,
                                                    name=gene2_contig.name,
                                                    greytrack=True,
                                                    greytrack_labels=1,
                                                    greytrack_font='Helvetica',
                                                    greytrack_fontsize=12,
                                                    height=0.35,
                                                    start=0,
                                                    end=len(gene2_contig),
                                                    scale=True,
                                                    scale_fontsize=6,
                                                    scale_ticks=1,
                                                    scale_smalltick_interval=10000,
                                                    scale_largetick_interval=10000)

    # add crosslink from blast results
    blast_results = open(path_to_blast_result)
    # parse blast results
    for each_line in blast_results:
        each_line_split = each_line.split('\t')
        query = each_line_split[0]
        target = each_line_split[1]
        identity = float(each_line_split[2])
        query_start = int(each_line_split[6])
        query_end = int(each_line_split[7])
        target_start = int(each_line_split[8])
        target_end = int(each_line_split[9])

        # use color to reflect identity
        color = colors.linearlyInterpolatedColor(colors.white, colors.red, 50, 100, identity)

        # determine which is which (query/target to contig_1/contig_2)
        # if query is contig_1
        if query == gene1_contig.name :
            link = CrossLink((contig_1_gene_content_track, query_start, query_end),
                             (contig_2_gene_content_track, target_start, target_end),
                             color = color,
                             border = color,
                             flip = False)
            diagram.cross_track_links.append(link)

        # if query is contig_2
        elif query == gene2_contig.name:
            link = CrossLink((contig_2_gene_content_track, query_start, query_end),
                             (contig_1_gene_content_track, target_start, target_end),
                             color = color,
                             border = color,
                             flip = False)
            diagram.cross_track_links.append(link)

    # Draw and Export
    diagram.draw(format="linear",
                 orientation="landscape",
                 pagesize=(75 * cm, 25 * cm),
                 fragments=1,
                 start=0,
                 end=max_len)

    diagram.write('%s.eps' % plot_file_name, "EPS")



import os
import glob
import shutil
import argparse
import subprocess
from time import sleep
from datetime import datetime
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC
from Bio.SeqRecord import SeqRecord


# to do
# include filter with blast coverage?


def export_dna_record(gene_seq, gene_id, gene_description, output_handle):
    seq_object = Seq(gene_seq, IUPAC.unambiguous_dna)
    seq_record = SeqRecord(seq_object)
    seq_record.id = gene_id
    seq_record.description = gene_description
    SeqIO.write(seq_record, output_handle, 'fasta')


def get_best_domain_hit_seq(pwd_domtblout_file, protein_Sequence_dict, pwd_output_folder):

    current_hmm_id = ''
    current_target_gene = ''
    current_hmm_pos1 = 0
    current_hmm_pos2 = 0
    current_hmm_score = 0
    # pwd_best_domain_hit_file_handle = open(pwd_best_domain_hit_file, 'w')
    for each_hit in open(pwd_domtblout_file):
        if not each_hit.startswith('#'):
            each_hit_split = each_hit.strip().split(' ')
            each_hit_split_no_space = []
            for each_element in each_hit_split:
                if each_element != '':
                    each_hit_split_no_space.append(each_element)
            target_gene = each_hit_split_no_space[0]
            hmm_id = each_hit_split_no_space[4]
            hmm_score = float(each_hit_split_no_space[13])
            hmm_pos1 = int(each_hit_split_no_space[17])
            hmm_pos2 = int(each_hit_split_no_space[18])

            if current_hmm_id == '':
                current_hmm_id = hmm_id
                current_target_gene = target_gene
                current_hmm_pos1 = hmm_pos1
                current_hmm_pos2 = hmm_pos2
                current_hmm_score = hmm_score
            elif current_hmm_id == hmm_id:
                if hmm_score > current_hmm_score:
                    current_hmm_id = hmm_id
                    current_target_gene = target_gene
                    current_hmm_pos1 = hmm_pos1
                    current_hmm_pos2 = hmm_pos2
                    current_hmm_score = hmm_score
            elif current_hmm_id != hmm_id:
                seqout_file_basename = '%s___%s' % (seqin_file_name, current_hmm_id)
                seqout_file_name = '%s.faa' % (seqout_file_basename)
                pwd_seqout_file_name = '%s/%s' % (pwd_output_folder, seqout_file_name)
                seqout_file_handle = open(pwd_seqout_file_name, 'w')
                export_dna_record(protein_Sequence_dict[current_target_gene][(current_hmm_pos1 - 1):current_hmm_pos2], seqout_file_basename, '', seqout_file_handle)
                seqout_file_handle.close()
                # pwd_best_domain_hit_file_handle.write('%s\t%s\t%s\t%s\t%s\n' % (current_hmm_id, current_target_gene, current_hmm_score, current_hmm_pos1, current_hmm_pos2))
                current_hmm_id = hmm_id
                current_target_gene = target_gene
                current_hmm_pos1 = hmm_pos1
                current_hmm_pos2 = hmm_pos2
                current_hmm_score = hmm_score
    # pwd_best_domain_hit_file_handle.write('%s\t%s\t%s\t%s\t%s\n' % (current_hmm_id, current_target_gene, current_hmm_score, current_hmm_pos1, current_hmm_pos2))
    # pwd_best_domain_hit_file_handle.close()


def get_assignment(list_in, confidence_level):

    # uniq list
    list_in_uniq = []
    for each_taxa in list_in:
        if each_taxa not in list_in_uniq:
            list_in_uniq.append(each_taxa)

    # get percentage for each uniq taxa and calculate assignment
    assignment = ''
    assignment_percentage = 0
    for each_uniq_taxa in list_in_uniq:
        each_uniq_taxa_count = list_in.count(each_uniq_taxa)
        each_uniq_taxa_percent = each_uniq_taxa_count * 100 / len(list_in)
        each_uniq_taxa_percent = float("{0:.2f}".format(each_uniq_taxa_percent))
        each_uniq_taxa_split = each_uniq_taxa.split('__')
        if (each_uniq_taxa_percent >= confidence_level) and (each_uniq_taxa_split[1] != ''):
            assignment = each_uniq_taxa
            assignment_percentage = each_uniq_taxa_percent

    return assignment, assignment_percentage


######################################################## inputs ########################################################

if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument('-i',
                        required=True,
                        help='folder holds the annotation file of input genomes')

    parser.add_argument('-c',
                        required=False,
                        type=int,
                        default=80,
                        help='confidence level (must higher than 50), default: 80')

    parser.add_argument('-tuning',
                        required=False,
                        action="store_true",
                        help='tuning mode')

    parser.add_argument('-bac120hmm',
                        required=True,
                        help='path to bac120.HMM')

    parser.add_argument('-blastdb',
                        required=True,
                        help='path to folder holds blast database')

    parser.add_argument('-taxondb',
                        required=True,
                        help='path to file bac_taxonomy_r83.tsv')

    parser.add_argument('-hmmsearch',
                        required=False,
                        default='hmmsearch',
                        help='path to hmmsearch, default: hmmsearch')

    parser.add_argument('-blastp',
                        required=False,
                        default='blastp',
                        help='path to blastp, default: blastp')

    parser.add_argument('-ktImportText',
                        required=False,
                        default='ktImportText',
                        help='path to ktImportText (from KronaTools), default: KronaTools')

    args = vars(parser.parse_args())
    genome_folder = args['i']
    confidence_level = args['c']
    pwd_hmm_db = args['bac120hmm']
    pwd_blastdb = args['blastdb']
    gtdb_taxonomy_db = args['taxondb']
    pwd_hmmsearch_exe = args['hmmsearch']
    pwd_blastp_exe = args['blastp']
    pwd_ktImportText_executable = args['ktImportText']
    tuning_mode = args['tuning']

    if genome_folder[-1] == '/':
        input_bin_folder_1 = genome_folder[:-1]

    #################################################### define file name ##################################################

    wd = os.getcwd()

    op_folder =                     '%s_TCG_outputs' % genome_folder.split('/')[-1]
    op_folder_temp =                'temp'
    op_folder_krona =               'Krona_plots'
    op_folder_krona_input_txt =     'files_for_Krona_plots'
    output_finest =                 'taxon_assignment_c%s.txt' % confidence_level
    output_lineage =                'taxon_assignment_lineage_c%s.txt' % confidence_level

    pwd_op_folder =                 '%s/%s'    % (wd, op_folder)
    pwd_op_folder_temp =            '%s/%s'    % (pwd_op_folder, op_folder_temp)
    pwd_op_folder_krona =           '%s/%s'    % (pwd_op_folder, op_folder_krona)
    pwd_op_folder_krona_input_txt = '%s/%s'    % (pwd_op_folder_krona, op_folder_krona_input_txt)
    pwd_output_finest =             '%s/%s'    % (pwd_op_folder, output_finest)
    pwd_output_lineage =            '%s/%s'    % (pwd_op_folder, output_lineage)

    # create temporary folder
    if tuning_mode != True:
        if os.path.isdir(pwd_op_folder):
            shutil.rmtree(pwd_op_folder, ignore_errors=True)
            if os.path.isdir(pwd_op_folder):
                shutil.rmtree(pwd_op_folder, ignore_errors=True)
            os.mkdir(pwd_op_folder)
            os.mkdir(pwd_op_folder_temp)
            os.mkdir(pwd_op_folder_krona)
            os.mkdir(pwd_op_folder_krona_input_txt)
        else:
            os.mkdir(pwd_op_folder)
            os.mkdir(pwd_op_folder_temp)
            os.mkdir(pwd_op_folder_krona)
            os.mkdir(pwd_op_folder_krona_input_txt)

    # get input genome file list
    input_genome_re = '%s/*.%s' % (genome_folder, 'faa')
    input_genome_file_list = [os.path.basename(file_name) for file_name in glob.glob(input_genome_re)]
    input_genome_file_list = sorted(input_genome_file_list)

    print()

    if len(input_genome_file_list) == 0:
        print(datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ' No input genome detected, program exited')
        exit()

    sleep(0.5)
    print(datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ' The number of input genome: %s' % (len(input_genome_file_list)))
    sleep(0.5)

    pwd_output_finest_handle = open(pwd_output_finest, 'w')
    pwd_output_finest_handle.write('Genome\tAssignment\tConfidence_level(%)\n')
    pwd_output_finest_handle.close()

    pwd_output_lineage_handle = open(pwd_output_lineage, 'w')
    pwd_output_lineage_handle.close()

    n = 1
    for each_query_genome in input_genome_file_list:
        pwd_query_genome = '%s/%s' % (genome_folder, each_query_genome)
        seqin_file_name, faa_file_ext = os.path.splitext(each_query_genome)

        print(datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ' Processing %s/%s: %s' % (n, len(input_genome_file_list), each_query_genome))

        ################################################## define file name ################################################

        output_subfolder =              '%s'                                % seqin_file_name
        domtblout_file =                '%s_hmmout.tsv'                     % seqin_file_name
        combined_blast_results =        '%s_combined_blast_results.txt'     % seqin_file_name
        identified_lineage =            '%s_identified_lineage.txt'         % seqin_file_name
        identified_lineage_sorted =     '%s_identified_lineage_sorted.txt'  % seqin_file_name
        krona_in =                      '%s_krona_in.txt'                   % seqin_file_name
        krona_out =                     'Krona_%s.html'                     % seqin_file_name
        krona_report =                  '%s_krona_report.txt'               % seqin_file_name

        pwd_output_subfolder =          '%s/%s'     % (pwd_op_folder_temp, output_subfolder)
        pwd_domtblout_file =            '%s/%s'     % (pwd_output_subfolder, domtblout_file)
        pwd_combined_blast_results =    '%s/%s'     % (pwd_output_subfolder, combined_blast_results)
        pwd_identified_lineage =        '%s/%s'     % (pwd_output_subfolder, identified_lineage)
        pwd_identified_lineage_sorted = '%s/%s'     % (pwd_output_subfolder, identified_lineage_sorted)
        pwd_krona_report =              '%s/%s'     % (pwd_output_subfolder, krona_report)
        pwd_krona_in =                  '%s/%s'     % (pwd_op_folder_krona_input_txt, krona_in)
        pwd_krona_out =                 '%s/%s'     % (pwd_op_folder_krona, krona_out)

        ####################################################################################################################

        # create temporary folder
        if tuning_mode != True:
            if os.path.isdir(pwd_output_subfolder):
                shutil.rmtree(pwd_output_subfolder)
                os.makedirs(pwd_output_subfolder)
            else:
                os.makedirs(pwd_output_subfolder)

        # get protein_Sequence_dict
        protein_Sequence_dict = {}
        for seq_record in SeqIO.parse(pwd_query_genome, 'fasta'):
            protein_Sequence_dict[seq_record.id] = str(seq_record.seq)

        # run hmmsearch
        if tuning_mode != True:
            hmmsearch_cmd = '%s -o /dev/null --domtblout %s %s %s' % (pwd_hmmsearch_exe, pwd_domtblout_file, pwd_hmm_db, pwd_query_genome)
            os.system(hmmsearch_cmd)

        # get best domain hit sequences
        if tuning_mode != True:
            get_best_domain_hit_seq(pwd_domtblout_file, protein_Sequence_dict, pwd_output_subfolder)

        # get file list
        file_re = '%s/*.%s' % (pwd_output_subfolder, 'faa')
        file_list = [os.path.basename(file_name) for file_name in glob.glob(file_re)]
        file_list = sorted(file_list)
        MaxProcesses = 20
        Processes = []

        for each_file in file_list:
            each_file_name, each_file_ext = os.path.splitext(each_file)
            hmm_id = each_file_name.split('___')[1]
            pwd_each_file = '%s/%s' % (pwd_output_subfolder, each_file)
            pwd_each_file_db = '%s/gtdb_%s.faa' % (pwd_blastdb, hmm_id)
            pwd_each_file_blastp_out = '%s/%s_blastp.txt' % (pwd_output_subfolder, each_file_name)
            blastp_cmd = '%s -query %s -subject %s -outfmt 6 -out %s -max_target_seqs 1; rm %s' % (pwd_blastp_exe, pwd_each_file, pwd_each_file_db, pwd_each_file_blastp_out, pwd_each_file)

            # run with subprocess
            blastp_cmd_list = [pwd_blastp_exe, '-query', pwd_each_file, '-subject', pwd_each_file_db, '-outfmt', '6', '-out', pwd_each_file_blastp_out, '-max_target_seqs', '1']
            if tuning_mode != True:
                # keep wait if there is no spare slots
                while len(Processes) >= MaxProcesses:
                    sleep(0.1)
                    for process in Processes:
                        if process.poll() is not None:
                            Processes.remove(process)
                # submit new subprocess
                p = subprocess.Popen(blastp_cmd_list)
                Processes.append(p)

        # wait for all subprocess to finish
        if tuning_mode != True:
            p.wait()

        # remove faa files
        if tuning_mode != True:
            os.system('rm %s/*.faa' % pwd_output_subfolder)

        # combine blast results together
        if tuning_mode != True:
            os.system('cat %s/*_blastp.txt > %s' % (pwd_output_subfolder, pwd_combined_blast_results))

        taxon_id_to_lineage_dict = {}
        for each_taxon in open(gtdb_taxonomy_db):
            each_taxon_split = each_taxon.strip().split('\t')
            each_taxon_id = each_taxon_split[0]
            each_taxon_lineage = each_taxon_split[1]
            taxon_id_to_lineage_dict[each_taxon_id] = each_taxon_lineage

        # get reference id list
        identified_lineage_handle = open(pwd_identified_lineage, 'w')
        for each in open(pwd_combined_blast_results):
            each_split = each.strip().split('\t')
            ref_id = each_split[1]
            identified_lineage_handle.write('%s\n' % (taxon_id_to_lineage_dict[ref_id]))
        identified_lineage_handle.close()

        ################################################## get Krona plot ##################################################

        # prepare input for Krona plot
        if tuning_mode != True:
            os.system('cat %s | sort > %s' % (pwd_identified_lineage, pwd_identified_lineage_sorted))

        identified_lineage_list = []
        identified_lineage_list_uniq = []
        for each_identified_lineage in open(pwd_identified_lineage_sorted):
            each_identified_lineage = each_identified_lineage.strip()
            identified_lineage_list.append(each_identified_lineage)
            if each_identified_lineage not in identified_lineage_list_uniq:
                identified_lineage_list_uniq.append(each_identified_lineage)

        identified_lineage_uniq_count_dict = {}
        for each_identified_lineage_uniq in identified_lineage_list_uniq:
            current_count = identified_lineage_list.count(each_identified_lineage_uniq)
            identified_lineage_uniq_count_dict[each_identified_lineage_uniq] = current_count

        # get input file for Krona plot
        pwd_krona_in_handle = open(pwd_krona_in, 'w')
        for each_key in identified_lineage_uniq_count_dict:
            each_key_split = each_key.split(';')
            pwd_krona_in_handle.write('%s\t%s\n' % (identified_lineage_uniq_count_dict[each_key], '\t'.join(each_key_split)))
        pwd_krona_in_handle.close()

        # plot Krona
        if tuning_mode != True:
            os.system('%s %s -o %s > %s' % (pwd_ktImportText_executable, pwd_krona_in, pwd_krona_out, pwd_krona_report))
            os.system('rm %s' % pwd_krona_report)

        ####################################################################################################################

        # get the number of assignment at all seven levels
        d_list = []
        p_list = []
        c_list = []
        o_list = []
        f_list = []
        g_list = []
        s_list = []
        for each_lineage in open(pwd_identified_lineage):
            each_lineage_split = each_lineage.strip().split(';')
            d_list.append(each_lineage_split[0])
            p_list.append(each_lineage_split[1])
            c_list.append(each_lineage_split[2])
            o_list.append(each_lineage_split[3])
            f_list.append(each_lineage_split[4])
            g_list.append(each_lineage_split[5])
            s_list.append(each_lineage_split[6])

        # put all list together for interation
        list_of_rank_list = [d_list, p_list, c_list, o_list, f_list, g_list, s_list]

        # get assignments at all rank levels
        assignment_list = []
        assignment_percentage_list = []
        assignment_combined = []
        for each_rank_list in list_of_rank_list:
            current_assignment, current_assignment_percentage = get_assignment(each_rank_list, confidence_level)
            if current_assignment_percentage >= confidence_level:
                assignment_list.append(current_assignment)
                assignment_percentage_list.append(current_assignment_percentage)
                combined = '%s(%s)' % (current_assignment, current_assignment_percentage)
                assignment_combined.append(combined)

        # get finest assignment
        finest_assignment = assignment_list[-1]
        finest_assignment_percent = assignment_percentage_list[-1]

        # write out
        pwd_output_finest_handle = open(pwd_output_finest, 'a')
        pwd_output_finest_handle.write('%s\t%s\t%s\n' % (seqin_file_name, finest_assignment, finest_assignment_percent))
        pwd_output_finest_handle.close()
        pwd_output_lineage_handle = open(pwd_output_lineage, 'a')
        pwd_output_lineage_handle.write('%s\t%s\n' % (seqin_file_name, ';'.join(assignment_combined)))
        pwd_output_lineage_handle.close()

        n += 1

    print(datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ' All done!')



#!/usr/bin/python
import os
import re
import argparse
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC
from Bio.SeqRecord import SeqRecord


def export_dna_record(gene_seq, gene_id, gene_description, output_handle):
    seq_object = Seq(gene_seq, IUPAC.unambiguous_dna)
    seq_record = SeqRecord(seq_object)
    seq_record.id = gene_id
    seq_record.description = gene_description
    SeqIO.write(seq_record, output_handle, 'fasta')


def treeMaker(path_to_prokka, path_to_hmm, pwd_hmmsearch_exe, pwd_mafft_exe, pwd_fasttree_exe, plot_tree):

    # Tests for presence of the tmp folder and deletes it
    tmp_folder = 'get_species_tree_wd'
    if os.path.exists(tmp_folder):
        os.system('rm -r ' + tmp_folder)
    os.mkdir(tmp_folder)

    # List all prokka dirs in the target folder
    prokka_files = [i for i in os.listdir(path_to_prokka) if os.path.isdir(path_to_prokka + '/' + i)]
    print('Detected %i input genomes' % len(prokka_files))

    # Running hmmsearch on each file
    print('Running hmmsearch...')
    for f in prokka_files:
        # call hmmsearch
        #os.system('hmmsearch -o /dev/null --domtblout %s/%s_hmmout.tbl %s %s/%s/%s.faa' % (tmp_folder, f, path_to_hmm, path_to_prokka, f, f))
        os.system('%s -o /dev/null --domtblout %s/%s_hmmout.tbl %s %s/%s/%s.faa' % (pwd_hmmsearch_exe, tmp_folder, f, path_to_hmm, path_to_prokka, f, f))

        # Reading the protein file in a dictionary
        proteinSequence = {}
        for seq_record in SeqIO.parse('%s/%s/%s.faa' % (path_to_prokka, f, f), 'fasta'):
            proteinSequence[seq_record.id] = str(seq_record.seq)

        # Reading the hmmersearch table/extracting the protein part found beu hmmsearch out of the protein/Writing each protein sequence that was extracted to a fasta file (one for each hmm in phylo.hmm
        hmm_id = ''
        hmm_name = ''
        hmm_pos1 = 0
        hmm_pos2 = 0
        hmm_score = 0

        with open(tmp_folder + '/' + f.replace('prokka/', '') + '_hmmout.tbl', 'r') as tbl:
            for line in tbl:
                if line[0] == "#": continue
                line = re.sub('\s+', ' ', line)
                splitLine = line.split(' ')

                if (hmm_id == ''):
                    hmm_id = splitLine[4]
                    hmm_name = splitLine[0]
                    hmm_pos1 = int(splitLine[17]) - 1
                    hmm_pos2 = int(splitLine[18])
                    hmm_score = float(splitLine[13])
                elif (hmm_id == splitLine[4]):
                    if(float(splitLine[13]) > hmm_score):
                        hmm_name = splitLine[0]
                        hmm_pos1 = int(splitLine[17]) - 1
                        hmm_pos2 = int(splitLine[18])
                        hmm_score = float(splitLine[13])
                else:
                    file_out = open(tmp_folder + '/' + hmm_id + '.fasta', 'a+')
                    file_out.write('>' + f + '\n')
                    if hmm_name != '':
                        seq = str(proteinSequence[hmm_name][hmm_pos1:hmm_pos2])
                    file_out.write(str(seq) + '\n')
                    file_out.close()
                    hmm_id = splitLine[4]
                    hmm_name = splitLine[0]
                    hmm_pos1 = int(splitLine[17]) - 1
                    hmm_pos2 = int(splitLine[18])
                    hmm_score = float(splitLine[13])

            else:
                file_out = open(tmp_folder + '/' + hmm_id + '.fasta', 'a+')
                file_out.write('>' + f + '\n')
                if hmm_name != '':
                    seq = str(proteinSequence[hmm_name][hmm_pos1:hmm_pos2])
                file_out.write(str(seq) + '\n')
                file_out.close()

    # Call mafft to align all single fasta files with hmms
    files = os.listdir(tmp_folder)
    fastaFiles = [i for i in files if i.endswith('.fasta')]
    print('Running mafft...')
    for f in fastaFiles:
        fastaFile1 = '%s/%s' % (tmp_folder, f)
        fastaFile2 = fastaFile1.replace('.fasta', '_aligned.fasta')
        os.system(pwd_mafft_exe + ' --quiet --maxiterate 1000 --globalpair ' + fastaFile1 + ' > ' + fastaFile2 + ' ; rm ' + fastaFile1)

    # concatenating the single alignments
    # create the dictionary
    print('Concatenating alignments...')
    concatAlignment = {}
    for element in prokka_files:
        concatAlignment[element] = ''

    # Reading all single alignment files and append them to the concatenated alignment
    files = os.listdir(tmp_folder)
    fastaFiles = [i for i in files if i.endswith('.fasta')]
    for f in fastaFiles:
        fastaFile = tmp_folder + '/' + f
        proteinSequence = {}
        alignmentLength = 0
        for seq_record_2 in SeqIO.parse(fastaFile, 'fasta'):
            proteinName =seq_record_2.id
            proteinSequence[proteinName] = str(seq_record_2.seq)
            alignmentLength = len(proteinSequence[proteinName])

        for element in prokka_files:
            if element in proteinSequence.keys():
                concatAlignment[element] += proteinSequence[element]
            else:
                concatAlignment[element] += '-' * alignmentLength

    # writing alignment to file
    file_out = open('./species_tree.aln', 'w')
    for element in prokka_files:
        file_out.write('>' + element + '\n' + concatAlignment[element] + '\n')
    file_out.close()

    # calling fasttree for tree calculation
    print('Running fasttree...')
    os.system('%s -quiet species_tree.aln > species_tree.newick' % pwd_fasttree_exe)

    # Decomment the two following lines if tree is rooted but should be unrooted
    #phyloTree = dendropy.Tree.get(path='phylogenticTree.phy', schema='newick', rooting='force-unrooted')
    #dendropy.Tree.write_to_path(phyloTree, 'phylogenticTree_unrooted.phy', 'newick')

    # plot species tree
    if plot_tree == 1:
        print('Plot species tree')

        tree = Tree('species_tree.newick', format=1)
        # set tree parameters
        ts = TreeStyle()
        ts.mode = "r"  # tree model: 'r' for rectangular, 'c' for circular
        ts.show_leaf_name = 0
        # set tree title text parameters
        ts.title.add_face(TextFace('Species_Tree',
                                   fsize=8,
                                   fgcolor='black',
                                   ftype='Arial',
                                   tight_text=False),
                          column=0)  # tree title text setting
        # set layout parameters
        ts.rotation = 0  # from 0 to 360
        ts.show_scale = False
        ts.margin_top = 10  # top tree image margin
        ts.margin_bottom = 10  # bottom tree image margin
        ts.margin_left = 10  # left tree image margin
        ts.margin_right = 10  # right tree image margin
        ts.show_border = False  # set tree image border
        ts.branch_vertical_margin = 3  # 3 pixels between adjancent branches

        # set tree node style
        for each_node in tree.traverse():
            # leaf node parameters
            if each_node.is_leaf():
                ns = NodeStyle()
                ns["shape"] = "circle"  # dot shape: circle, square or sphere
                ns["size"] = 0  # dot size
                ns['hz_line_width'] = 0.5  # branch line width
                ns['vt_line_width'] = 0.5  # branch line width
                ns['hz_line_type'] = 0  # branch line type: 0 for solid, 1 for dashed, 2 for dotted
                ns['vt_line_type'] = 0  # branch line type
                ns["fgcolor"] = "blue"  # the dot setting
                each_node.add_face(TextFace(each_node.name,
                                            fsize=5,
                                            fgcolor='black',
                                            tight_text=False,
                                            bold=False),
                                   column=0,
                                   position='branch-right')  # leaf node the node name text setting

                each_node.set_style(ns)

            # non-leaf node parameters
            else:
                nlns = NodeStyle()
                nlns["size"] = 0  # dot size
                # nlns["rotation"] = 45
                each_node.add_face(TextFace(each_node.name,
                                            fsize=3,
                                            fgcolor='black',
                                            tight_text=False,
                                            bold=False),
                                   column=5,
                                   position='branch-top')  # non-leaf node name text setting)

                each_node.set_style(nlns)

        tree.render('species_tree' + '.png', w=900, units="px", tree_style=ts)  # set figures size

    if plot_tree == 0:
        print('The built species tree was exported to species_tree.newick')
    else:
        print('The built species tree was exported to species_tree.newick and species_tree.png')


parser = argparse.ArgumentParser()

parser.add_argument('-prokka_output',
                    required=True,
                    help='the folder holds Prokka outputs for input genomes')

parser.add_argument('-hmm',
                    required=True,
                    help='the phylo.hmm file')

parser.add_argument('-hmmsearch',
                    required=False,
                    default='hmmsearch',
                    help='path to hmmsearch executable')

parser.add_argument('-mafft',
                    required=False,
                    default='mafft',
                    help='path to mafft executable')

parser.add_argument('-fasttree',
                    required=False,
                    default='fasttree',
                    help='path to fasttree executable')

parser.add_argument('-plot_tree',
                    required=False,
                    action='store_true',
                    help='to plot species tree')

args = vars(parser.parse_args())
prokka_output = args['prokka_output']
hmm_file = args['hmm']
pwd_hmmsearch_exe = args['hmmsearch']
pwd_mafft_exe = args['mafft']
pwd_fasttree_exe = args['fasttree']
plot_tree = args['plot_tree']


if plot_tree == 1:
    from ete3 import Tree, TreeStyle, NodeStyle, TextFace

treeMaker(prokka_output, hmm_file, pwd_hmmsearch_exe, pwd_mafft_exe, pwd_fasttree_exe, plot_tree)



import os
from ete3 import Tree
from ete3 import TreeStyle
from ete3 import NodeStyle
from ete3 import TextFace


def plot_tree(tree, tree_title, tree_output):
    # set tree parameters
    ts = TreeStyle()
    ts.mode = "r"  # tree model: 'r' for rectangular, 'c' for circular
    ts.show_leaf_name = 0
    # set tree title text parameters
    ts.title.add_face(TextFace(tree_title,
                               fsize = 8,
                               fgcolor = 'black',
                               ftype = 'Arial',
                               tight_text = False),
                      column = 0)  # tree title text setting
    # set layout parameters
    ts.rotation = 0  # from 0 to 360
    ts.show_scale = False
    ts.margin_top = 10  # top tree image margin
    ts.margin_bottom = 10  # bottom tree image margin
    ts.margin_left = 10  # left tree image margin
    ts.margin_right = 10  # right tree image margin
    ts.show_border = False  # set tree image border
    ts.branch_vertical_margin = 3  # 3 pixels between adjancent branches

    # set tree node style
    for each_node in tree.traverse():
        # leaf node parameters
        if each_node.is_leaf():
            ns = NodeStyle()
            ns["shape"] = "circle"  # dot shape: circle, square or sphere
            ns["size"] = 0  # dot size
            ns['hz_line_width'] = 0.5  # branch line width
            ns['vt_line_width'] = 0.5  # branch line width
            ns['hz_line_type'] = 0  # branch line type: 0 for solid, 1 for dashed, 2 for dotted
            ns['vt_line_type'] = 0  # branch line type
            ns["fgcolor"] = "blue"  # the dot setting
            each_node.add_face(TextFace(each_node.name,
                                        fsize = 5,
                                        fgcolor = 'black',
                                        tight_text = False,
                                        bold = False),
                               column = 0,
                               position = 'branch-right')  # leaf node the node name text setting

            each_node.set_style(ns)

        # non-leaf node parameters
        else:
            nlns = NodeStyle()
            nlns["size"] = 0  # dot size
            #nlns["rotation"] = 45
            each_node.add_face(TextFace(each_node.name,

                                        fsize = 3,
                                        fgcolor = 'black',
                                        tight_text = False,
                                        bold = False),
                               column = 5,
                               position = 'branch-top')  # non-leaf node name text setting)

            each_node.set_style(nlns)

    tree.render(tree_output, w=900, units="px", tree_style=ts)  # set figures size


os.chdir('/Users/songweizhi/Desktop')

tree = Tree('/Users/songweizhi/Desktop/species_tree.newick', format = 0)
plot_tree(tree, 'Overall', 'Overall.png')

sub_tree = Tree('/Users/songweizhi/Desktop/species_tree.newick', format = 0)
sub_tree.prune(['AMS', 'AKV', 'BDS', 'BAD'])
plot_tree(sub_tree, 'sub_tree', 'sub_tree.png')



import os
import shutil

def force_create_folder(folder_to_create):
    if os.path.isdir(folder_to_create):
        shutil.rmtree(folder_to_create, ignore_errors=True)
        if os.path.isdir(folder_to_create):
            shutil.rmtree(folder_to_create, ignore_errors=True)
            if os.path.isdir(folder_to_create):
                shutil.rmtree(folder_to_create, ignore_errors=True)
                if os.path.isdir(folder_to_create):
                    shutil.rmtree(folder_to_create, ignore_errors=True)
    os.mkdir(folder_to_create)


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


########################################################################################################################

combined_checkm_quality_file = '/Users/songweizhi/Desktop/Mariana_Trench_checkm.txt'
completeness_cutoff = 50
contamination_cutoff = 5
bin_folder_in = '/Users/songweizhi/Desktop/Mariana_Trench_refined_bins_all'


########################################################################################################################


infile_path, infile_basename, infile_extension = sep_path_basename_ext(combined_checkm_quality_file)
file_out = '%s/%s_cmplt%s_contam%s%s' % (infile_path, infile_basename, str(completeness_cutoff), str(contamination_cutoff), infile_extension)


bin_folder_out = '%s_%s_%s' % (bin_folder_in, completeness_cutoff, contamination_cutoff)
force_create_folder(bin_folder_out)


filtered_checkm_quality_handle = open(file_out, 'w')
filtered_checkm_quality_handle.write('Genome_ID\tCompleteness\tContamination\n')
qualified_bin_num = 0
for each in open(combined_checkm_quality_file):

    if not ((each.startswith('---')) or (each.startswith('  Bin Id'))):
        quality_split = each.strip().split(' ')
        quality_split_new = []
        for each in quality_split:
            if each != '':
                quality_split_new.append(each)

        genome_id = quality_split_new[0]
        completeness = float(quality_split_new[12])
        contamination = float(quality_split_new[13])

        if (completeness >= completeness_cutoff) and (contamination <= contamination_cutoff):
            filtered_checkm_quality_handle.write('%s\t%s\t%s\n' % (genome_id, completeness, contamination))
            #os.system('cp %s/%s.fasta %s/' % (bin_folder_in, genome_id, bin_folder_out))
            qualified_bin_num += 1

filtered_checkm_quality_handle.close()

print('The number of bins with completeness >= %s and contamination <= %s: %s' % (completeness_cutoff, contamination_cutoff, qualified_bin_num))

